#-------------------------------------------------------------------------------
#
#  WPS specific utilities.
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from contextlib import closing
from logging import getLogger

try:
    # available in Python 2.7+
    from collections import OrderedDict
except ImportError:
    from django.utils.datastructures import SortedDict as OrderedDict

from django.conf import settings
from django.utils.module_loading import import_string
from django.utils.six.moves.urllib import parse, request, error

from eoxserver.core.util.multiparttools import iterate as iterate_multipart
from eoxserver.services.ows.wps.config import (
    DEFAULT_EOXS_PROCESSES, DEFAULT_EOXS_ASYNC_BACKENDS
)


def parse_named_parts(request):
    """ Extract named parts of the multi-part request
    and return them as dictionary
    """
    parts = {}
    if request.method == 'POST':
        content_type = request.META.get("CONTENT_TYPE", "")
        if content_type.startswith("multipart"):
            parts = dict(
                (content_id, data) for content_id, data in (
                    (headers.get("Content-Id"), data)
                    for headers, data in iterate_multipart(
                        request.body, headers={"Content-Type": content_type}
                    )
                ) if content_id
            )
    return parts


class InMemoryURLResolver(object):
    # pylint: disable=too-few-public-methods, no-self-use
    """ Simple in-memory URL resolver.
    The resolver resolves references and returns them as data strings.
    """

    def __init__(self, parts=None, logger=None):
        self.parts = parts or {}
        self.logger = logger or getLogger(__name__)

    def __call__(self, href, body, headers):
        """ Resolve reference URL. """
        self.logger.debug(
            "Resolving reference: %s%s", href, "" if body is None else " (POST)"
        )
        url = parse.urlparse(href)
        if url.scheme == "cid":
            return self._resolve_multipart(url.path)
        elif url.scheme in ('http', 'https'):
            return self._resolve_http(href, body, headers)
        else:
            raise ValueError("Unsupported URL scheme %r!" % url.scheme)

    def _resolve_multipart(self, content_id):
        """ Resolve multipart-related."""
        try:
            return self.parts[content_id]
        except KeyError:
            raise ValueError("No part with content-id %r." % content_id)

    def _resolve_http(self, href, body=None, headers=None):
        """ Resolve the HTTP request."""
        try:
            with closing(request.urlopen(request.Request(href, body, dict(headers)))) as fobj:
                return fobj.read()
        except error.URLError as exc:
            raise ValueError(str(exc))


PROCESSES = None
ASYNC_BACKENDS = None

def _setup_processes():
    global PROCESSES
    specifiers = getattr(
        settings, 'EOXS_PROCESSES',
        DEFAULT_EOXS_PROCESSES
    )
    PROCESSES = [
        import_string(identifier)()
        for identifier in specifiers
    ]

def _setup_async_backends():
    global ASYNC_BACKENDS
    specifiers = getattr(
        settings, 'EOXS_ASYNC_BACKENDS',
        DEFAULT_EOXS_ASYNC_BACKENDS
    )
    ASYNC_BACKENDS = [
        import_string(identifier)()
        for identifier in specifiers
    ]


def get_processes():
    if PROCESSES is None:
        _setup_processes()

    return PROCESSES


def get_async_backends():
    if ASYNC_BACKENDS is None:
        _setup_async_backends()

    return ASYNC_BACKENDS



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class AsyncBackendInterface(object):
    """ Interface class for an asynchronous WPS back-end.
        NOTE: Only one asynchronous back-end at time is allowed to be configured.
    """
    @property
    def supported_versions(self):
        """ A list of versions of the WPS standard supported by the back-end.
        """

    def execute(self, process, raw_inputs, resp_form, extra_parts=None,
                job_id=None, version="1.0.0", **kwargs):
        """ Execute process asynchronously.
        The request is defined by the process's identifier ``process_id``,
        ``raw_inputs`` (before the decoding and resolution
        of the references), and the ``resp_form`` (holding
        the outputs' parameters).  The ``version`` of the WPS standard
        to be used.  Optionally, the user defined ``job_id`` can be passed.
        If the ``job_id`` cannot be used the execute shall fail.

        The ``extra_parts`` should contain a dictionary of named request parts
        should the request contain multi-part/related CID references.

        On success, the method returns the ``job_id`` assigned to the
        executed job.
        """

    def get_response_url(self, job_id):
        """ Get URL of the execute response for the given job id """

    def get_status(self, job_id):
        """ Get status of a job. Allowed responses and their meanings are:
            ACCEPTED  - job scheduled for execution
            STARTED   - job in progress
            PAUSED    - job is stopped and it can be resumed
            CANCELLED - job was terminated by the user
            FAILED    - job ended with an error
            SUCCEEDED - job ended successfully
        """

    def purge(self, job_id, **kwargs):
        """ Purge the job from the system by removing all the resources
        occupied by the job.
        """

    def cancel(self, job_id, **kwargs):
        """ Cancel the job execution. """

    def pause(self, job_id, **kwargs):
        """ Pause the job execution. """

    def resume(self, job_id, **kwargs):
        """ Resume the job execution. """


class ProcessInterface(object):
    """ Interface class for processes offered, described and executed by
        the WPS.
    """

    @property
    def version(self):
        """ The version of the process, if applicable. Optional.
            When omitted it defaults to '1.0.0'.
        """

    @property
    def synchronous(self):
        """ Optional boolean flag indicating whether the process can be executed
        synchronously. If missing True is assumed.
        """

    @property
    def asynchronous(self):
        """ Optional boolean flag indicating whether the process can be executed
        asynchronously. If missing False is assumed.
        """

    @property
    def retention_period(self):
        """ This optional property (`datetime.timedelta`) indicates the minimum
        time the process results shall be retained after the completion.
        If omitted the default server retention policy is applied.
        """

    @property
    def identifier(self):
        """ An identifier (URI) of the process. Optional.
            When omitted it defaults to the process' class-name.
        """

    @property
    def title(self):
        """ A human-readable title of the process. Optional. When omitted it
            defaults to the process identifier.
        """

    @property
    def description(self):
        """ A human-readable detailed description of the process. Optional.
            (Content of the the abstract in the WPS process description.)
        """

    @property
    def profiles(self):
        """ A iterable of URNs of WPS application profiles this process
            adheres to. Optional.
        """

    @property
    def metadata(self):
        """ A dict of title/URL meta-data pairs associated with the process.
            Optional.
        """

    @property
    def wsdl(self):
        """ A URL of WSDL document describing this process. Optional.
        """

    @property
    def inputs(self):
        """ A dict mapping the inputs' identifiers to their respective types.
            The type can be either one of the supported native python types
            (automatically converted to a ``LiteralData`` object) or an instance
            of one of the data-specification classes (``LiteralData``,
            ``BoundingBoxData``, or ``ComplexData``).  Mandatory.
        """

    @property
    def outputs(self):
        """ A dict mapping the outputs' identifiers to their respective types.
            The type can be either one of the supported native python types
            (automatically converted to a ``LiteralData`` object) or an instance
            of one of the data-specification classes (``LiteralData``,
            ``BoundingBoxData``, or ``ComplexData``).  Mandatory.
        """

    def execute(self, **kwargs):
        """ The main execution function for the process. The ``kwargs`` are the
            parsed input inputs (using the keys as defined by the ``inputs``)
            and the Complex Data format requests (using the keys as defined by
            the ``outputs``).
            The method is expected to return a dictionary of the output values
            (using the keys as defined by the ``outputs``). In case of only
            one output item defined by the ``outputs``, one output value
            is allowed to be returned directly.
        """



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""This model contains Django views for the EOxServer software. Its main
function is :func:`ows` which handles all incoming OWS requests"""

import logging
import traceback

from django.http import HttpResponse
try:
    from django.http import StreamingHttpResponse
except:
    class StreamingHttpResponse(object):
        pass
from django.views.decorators.csrf import csrf_exempt
from django.utils.six import string_types

from eoxserver.core import env
from eoxserver.services.ows.component import ServiceComponent
from eoxserver.services.exceptions import HTTPMethodNotAllowedError
from eoxserver.services.ows.dispatch import (
    query_service_handler, query_exception_handler
)


logger = logging.getLogger(__name__)


@csrf_exempt
def ows(request):
    """ Main entry point for OWS requests against EOxServer. It uses the
    :class:`ServiceComponent
    <eoxserver.services.ows.component.ServiceComponent>` to dynamically
    determine the handler component for this request.

    If an exception occurs during the handling of the request, an exception
    handler component is determined and dispatched.

    Any response of the service handler and exception handler is transformed
    to a django :class:`HttpResponse <django.http.HttpResponse>` to adhere the
    required interface.
    """

    # component = ServiceComponent(env)

    try:
        handler = query_service_handler(request)
        result = handler.handle(request)
        default_status = 200
    except HTTPMethodNotAllowedError as e:
        handler = query_exception_handler(request)
        result = handler.handle_exception(request, e)
        content, content_type = handler.handle_exception(request, e)[:2]
        result = HttpResponse(
            content=content, content_type=content_type, status=405
        )
        result["Allow"] = ", ".join(e.allowed_methods)
    except Exception as e:
        logger.debug(traceback.format_exc())
        handler = query_exception_handler(request)
        result = handler.handle_exception(request, e)
        default_status = 400

    # try to return a django compatible response
    if isinstance(result, (HttpResponse, StreamingHttpResponse)):
        return result

    elif isinstance(result, string_types):
        return HttpResponse(result)

    # convert result to a django response
    try:
        content, content_type, status = result
        return HttpResponse(
            content=content, content_type=content_type, status=status
        )
    except ValueError:
        pass

    try:
        content, content_type = result
        return HttpResponse(
            content=content, content_type=content_type, status=default_status
        )
    except ValueError:
        pass



#-------------------------------------------------------------------------------
#
# Base context class
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class ContextError(Exception):
    """ General context error. """
    pass


class BaseContext(object):
    """ Base context class.
    This is the base abstract asynchronous WPS context class defining the
    minimal common interface of a context object passed to an asynchronous
    WPS process.

    The purpose of the context is to manage status, storage and other resources
    used by the job (executed process instance).

    The actual implementation of the concrete context classes is part
    of the asynchronous implementation.
    """

    @property
    def identifier(self):
        """ Get the context specific identifier (job id.)
        """
        raise NotImplementedError

    @property
    def logger(self):
        """ Get the context specific logger. The returned logger is expected
        to be a context aware logger adapter.
        """
        raise NotImplementedError

    @property
    def workspace_path(self):
        """ Get the workspace path.
        Note that this path is expected to be the working directory when
        an asynchronous process gets executed. This allows the process
        to write safely to its working directory.
        """
        raise NotImplementedError

    def publish(self, path):
        """ Publish file from the local workspace and return its path
        and public URL.
        The input file path must be relative to the workspace path.
        Publishing of files outside of the workspace directory sub-tree
        is not allowed.
        """
        raise NotImplementedError

    def update_progress(self, progress, message):
        """ Update the StatusStarted response and set the new progress
        expressed in percent.  The progress therefore must be a number
        between 0 and 99.  An optional message can be specified.
        """
        raise NotImplementedError



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import operator

from django.db.models import Min, Max
from django.contrib.gis.db.models import Union
from django.contrib.gis.geos import MultiPolygon, Polygon
from django.utils.timezone import is_naive, make_aware, get_current_timezone
from django.utils.dateparse import parse_datetime
from django.utils.six import string_types

from eoxserver.contrib import gdal


def pk_equals(first, second):
    """ Helper function to check if the ``pk`` attributes of two models are
    equal.
    """
    return first.pk == second.pk


def collect_eo_metadata(qs, insert=None, exclude=None, bbox=False):
    """ Helper function to collect EO metadata from all EOObjects in a queryset,
    plus additionals from a list and exclude others from a different list. If
    bbox is ``True`` then the returned polygon will only be a minimal bounding
    box of the collected footprints.

    :param qs: the :class:`django.db.QuerySet` to collect all EO metadata from
    :param insert: an iterable of all objects that are to be inserted (thus not
                   entailed in the queryset) and should be considered when
                   collection metadata
    :param exclude: an iterable of objects that are considered to be excluded
                    from the list and should not be used for metadata collection
    :param bbox: if this is set to ``True`` the footprint will only be
                 represented as a minimal BBOX polygon of all collected
                 footprints. This is preferable for large collections.
    """

    values = qs.exclude(
        pk__in=[eo_object.pk for eo_object in exclude or ()]
    ).aggregate(
        begin_time=Min("begin_time"), end_time=Max("end_time"),
        footprint=Union("footprint")
    )

    begin_time, end_time, footprint = (
        values["begin_time"], values["end_time"], values["footprint"]
    )

    # workaround for Django 1.4 bug: aggregate times are strings
    if isinstance(begin_time, string_types):
        begin_time = parse_datetime(begin_time)

    if isinstance(end_time, string_types):
        end_time = parse_datetime(end_time)

    if begin_time and is_naive(begin_time):
        begin_time = make_aware(begin_time, get_current_timezone())
    if end_time and is_naive(end_time):
        end_time = make_aware(end_time, get_current_timezone())

    for eo_object in insert or ():
        if begin_time is None:
            begin_time = eo_object.begin_time
        elif eo_object.begin_time is not None:
            begin_time = min(begin_time, eo_object.begin_time)

        if end_time is None:
            end_time = eo_object.end_time
        elif eo_object.end_time is not None:
            end_time = max(end_time, eo_object.end_time)

        if footprint is None:
            footprint = eo_object.footprint
        elif eo_object.footprint is not None:
            footprint = footprint.union(eo_object.footprint)

    if not isinstance(footprint, MultiPolygon) and footprint is not None:
        footprint = MultiPolygon(footprint)

    if bbox and footprint is not None:
        footprint = MultiPolygon(Polygon.from_bbox(footprint.extent))

    return begin_time, end_time, footprint


def is_same_grid(coverages, epsilon=1e-10):
    """ Function to determine if the given coverages share the same base grid.
    Returns a boolean value, whether or not the coverages share a common grid.

    :param coverages: an iterable of :class:`Coverages
                      <eoxserver.resources.coverages.models.Coverage>`
    :param epsilon: the maximum difference each grid can have to be still
                    considered equal
    """

    if len(coverages) < 2:
        raise ValueError("Not enough coverages given.")

    first = coverages[0]
    first_ext = first.extent
    first_res = first.resolution
    first_srid = first.srid
    first_proj = first.projection

    e = epsilon

    for other in coverages[1:]:
        other_ext = other.extent
        other_res = other.resolution

        # check projection/srid
        if first_srid != other.srid or first_proj != other.projection:
            return False

        # check dimensions
        if len(first_res) != len(other_res):
            return False

        # check offset vectors
        for a, b in zip(first_res, other_res):
            if abs(a - b) > epsilon:
                return False

        # check base grids
        diff_origins = tuple(map(operator.sub, first_ext[:2], other_ext[:2]))

        v = tuple(map(operator.div, diff_origins, other_res))

        if (abs(v[0] - round(v[0])) > e or abs(v[1] - round(v[1])) > e):
            return False

    return True


def parse_raw_value(raw_value, dt):
    """ Parse a raw value from a string according to a given data type.

    :param raw_value: the raw string value
    :param dt: the data type enumeration
    :returns: the parsed value
    """
    if raw_value is None or raw_value == "":  # allow null and empty values
        return None

    is_float = False
    is_complex = False

    if dt in gdal.GDT_INTEGRAL_TYPES:
        value = int(raw_value)

    elif dt in gdal.GDT_FLOAT_TYPES:
        value = float(raw_value)
        is_float = True

    elif dt in gdal.GDT_INTEGRAL_COMPLEX_TYPES:
        value = complex(raw_value)
        is_complex = True

    elif dt in gdal.GDT_FLOAT_COMPLEX_TYPES:
        value = complex(raw_value)
        is_complex = True
        is_float = True

    else:
        value = None

    # range check makes sense for integral values only
    if not is_float:
        limits = gdal.GDT_NUMERIC_LIMITS.get(dt)

        if limits and value is not None:
            def within(v, low, high):
                return (v >= low and v <= high)

            error = ValueError(
                "Stored value is out of the limits for the data type"
            )
            if not is_complex and not within(value, *limits):
                raise error
            elif is_complex:
                if (not within(value.real, limits[0].real, limits[1].real)
                    or not within(value.real, limits[0].real, limits[1].real)):
                    raise error

    return value



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=missing-docstring

from sys import stdin
import traceback
import json
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.utils.six import string_types
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, nested_commit_on_success,
)
from eoxserver.resources.coverages.models import Coverage
from eoxserver.services.models import WMSRenderOptions


class Command(CommandOutputMixIn, BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option(
            '-i', '--input', dest='filename', action='store', type='string',
            default='-', help=(
                "Optional. Read input from a file rather than from the "
                "default standard input."
            )
        ),
    )

    help = """
    Load WMS options stored in JSON format from standard input (default) or from
    a file (-i option).
    """

    def handle(self, *args, **options):
        # Collect parameters
        self.traceback = bool(options.get("traceback", False))
        self.verbosity = int(options.get('verbosity', 1))
        filename = options.get('filename', '-')

        self.print_msg("Importing WMS options from %s ..." % (
            "standard input" if filename == "-" else "file %r" % filename
        ))

        # load and parse the input data
        try:
            with (stdin if filename == "-" else open(filename, "r")) as fin:
                wms_opts = json.load(fin)
        except IOError as exc:
            raise CommandError(
                "Failed to open the input file '%s'! %s " % (filename, str(exc))
            )

        # insert the range types to DB

        success_count = 0  # success counter - counts finished syncs

        for wms_opt in wms_opts:
            try:
                eoid = wms_opt['coverage_identifier']
                insert_or_update_wms_opt(eoid, wms_opt)
            except Exception as exc:
                if self.traceback:
                    self.print_msg(traceback.format_exc())
                self.print_err("Failed to load options of %s! %s" % (
                    eoid, "%s: %s" % (type(exc).__name__, str(exc))
                ))
                continue
            else:
                success_count += 1  # increment success counter

        # print the final summary
        count = len(wms_opts)
        error_count = count - success_count

        if success_count > 0:
            self.print_msg(
                "Successfully loaded %d of %s WMS option records." %
                (success_count, count), 1
            )
        else:
            self.print_msg("No WMS option record loaded.")

        if error_count > 0:
            raise CommandError(
                "Failed to load %d WMS option records." % error_count
            )


@nested_commit_on_success
def insert_or_update_wms_opt(eoid, wms_opt):
    """ Insert or update one WMS option record. """
    try:
        # try to get an existing object
        wms_opt_obj = WMSRenderOptions.objects.get(coverage__identifier=eoid)
    except WMSRenderOptions.DoesNotExist:
        try:
            # locate the related coverage
            coverage = Coverage.objects.get(identifier=eoid)
        except Coverage.DoesNotExist:
            raise CommandError("Invalid coverage identifier %s!" % eoid)
        # create a new object
        wms_opt_obj = WMSRenderOptions(coverage=coverage)

    # set the record
    set_from_dict(wms_opt_obj, wms_opt).save()


def set_from_dict(wms_opt_obj, options):
    """ Set object from a JSON serializable dictionary. """

    _int = lambda v: v if v is None else int(v)

    wms_opt_obj.default_red = _int(options.get('red', None))
    wms_opt_obj.default_green = _int(options.get('green', None))
    wms_opt_obj.default_blue = _int(options.get('blue', None))
    wms_opt_obj.default_alpha = _int(options.get('alpha', None))
    wms_opt_obj.resampling = options.get('resampling', None)
    wms_opt_obj.scale_auto = options.get('scale_auto', False)
    wms_opt_obj.scale_min = _int(options.get('scale_min', None))
    wms_opt_obj.scale_max = _int(options.get('scale_max', None))

    def _scales(scales):
        if scales is None:
            return None
        if isinstance(scales, string_types):
            scales = scales.split(',')
        return ",".join(str(int(round(float(v)))) for v in scales)

    wms_opt_obj.bands_scale_min = _scales(options.get('bands_scale_min', None))
    wms_opt_obj.bands_scale_max = _scales(options.get('bands_scale_max', None))

    return wms_opt_obj



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from itertools import chain

from eoxserver.core.decoders import xml, kvp, typelist
from eoxserver.services.subset import Subsets
from eoxserver.services.ows.wcs.basehandlers import WCSGetCoverageHandlerBase
from eoxserver.services.ows.wcs.v20.util import (
    nsmap, parse_subset_kvp, parse_subset_xml, parse_range_subset_kvp,
    parse_range_subset_xml, parse_interpolation,
    parse_scaleaxis_kvp, parse_scalesize_kvp, parse_scaleextent_kvp,
    parse_scaleaxis_xml, parse_scalesize_xml, parse_scaleextent_xml,
)
from eoxserver.services.ows.wcs.v20.parameters import WCS20CoverageRenderParams
from eoxserver.services.ows.wcs.v20.encodings import get_encoding_extensions
from eoxserver.services.exceptions import InvalidRequestException


class WCS20GetCoverageHandler(WCSGetCoverageHandlerBase):
    versions = ("2.0.0", "2.0.1")
    methods = ['GET', 'POST']

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS20GetCoverageKVPDecoder(request.GET)
        elif request.method == "POST":
            return WCS20GetCoverageXMLDecoder(request.body)

    def get_subsets(self, decoder):
        return Subsets(decoder.subsets, crs=decoder.subsettingcrs)

    def get_params(self, coverage, decoder, request):
        subsets = Subsets(decoder.subsets, crs=decoder.subsettingcrs)
        encoding_params = None
        for encoding_extension in get_encoding_extensions():
            if encoding_extension.supports(decoder.format, {}):
                encoding_params = encoding_extension.get_encoding_params(
                    request
                )

        scalefactor = decoder.scalefactor
        scales = list(
            chain(decoder.scaleaxes, decoder.scalesize, decoder.scaleextent)
        )

        # check scales validity: ScaleFactor and any other scale
        if scalefactor and scales:
            raise InvalidRequestException(
                "ScaleFactor and any other scale operation are mutually "
                "exclusive.", locator="scalefactor"
            )

        # check scales validity: Axis uniqueness
        axes = set()
        for scale in scales:
            if scale.axis in axes:
                raise InvalidRequestException(
                    "Axis '%s' is scaled multiple times." % scale.axis,
                    locator=scale.axis
                )
            axes.add(scale.axis)

        return WCS20CoverageRenderParams(
            coverage, subsets, decoder.rangesubset, decoder.format,
            decoder.outputcrs, decoder.mediatype, decoder.interpolation,
            scalefactor, scales, encoding_params or {}, request
        )


class WCS20GetCoverageKVPDecoder(kvp.Decoder):
    coverage_id = kvp.Parameter("coverageid", num=1)
    subsets     = kvp.Parameter("subset", type=parse_subset_kvp, num="*")
    scalefactor = kvp.Parameter("scalefactor", type=float, num="?")
    scaleaxes   = kvp.Parameter("scaleaxes", type=typelist(parse_scaleaxis_kvp, ","), default=(), num="?")
    scalesize   = kvp.Parameter("scalesize", type=typelist(parse_scalesize_kvp, ","), default=(), num="?")
    scaleextent = kvp.Parameter("scaleextent", type=typelist(parse_scaleextent_kvp, ","), default=(), num="?")
    rangesubset = kvp.Parameter("rangesubset", type=parse_range_subset_kvp, num="?")
    format      = kvp.Parameter("format", num="?")
    subsettingcrs = kvp.Parameter("subsettingcrs", num="?")
    outputcrs   = kvp.Parameter("outputcrs", num="?")
    mediatype   = kvp.Parameter("mediatype", num="?")
    interpolation = kvp.Parameter("interpolation", type=parse_interpolation, num="?")


class WCS20GetCoverageXMLDecoder(xml.Decoder):
    coverage_id = xml.Parameter("wcs:CoverageId/text()", num=1, locator="coverageid")
    subsets     = xml.Parameter("wcs:DimensionTrim", type=parse_subset_xml, num="*", locator="subset")
    scalefactor = xml.Parameter("wcs:Extension/scal:ScaleByFactor/scal:scaleFactor/text()", type=float, num="?", locator="scalefactor")
    scaleaxes   = xml.Parameter("wcs:Extension/scal:ScaleByAxesFactor/scal:ScaleAxis", type=parse_scaleaxis_xml, num="*", default=(), locator="scaleaxes")
    scalesize   = xml.Parameter("wcs:Extension/scal:ScaleToSize/scal:TargetAxisSize", type=parse_scalesize_xml, num="*", default=(), locator="scalesize")
    scaleextent = xml.Parameter("wcs:Extension/scal:ScaleToExtent/scal:TargetAxisExtent", type=parse_scaleextent_xml, num="*", default=(), locator="scaleextent")
    rangesubset = xml.Parameter("wcs:Extension/rsub:RangeSubset", type=parse_range_subset_xml, num="?", locator="rangesubset")
    format      = xml.Parameter("wcs:format/text()", num="?", locator="format")
    subsettingcrs = xml.Parameter("wcs:Extension/crs:subsettingCrs/text()", num="?", locator="subsettingcrs")
    outputcrs   = xml.Parameter("wcs:Extension/crs:outputCrs/text()", num="?", locator="outputcrs")
    mediatype   = xml.Parameter("wcs:mediaType/text()", num="?", locator="mediatype")
    interpolation = xml.Parameter("wcs:Extension/int:Interpolation/int:globalInterpolation/text()", type=parse_interpolation, num="?", locator="interpolation")

    namespaces = nsmap



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import logging

from django.conf import settings
# from django.utils.module_loading import import_string
from django.http import HttpResponse
from django.utils.six import string_types

from eoxserver.services.ows.config import (
    DEFAULT_EOXS_OWS_SERVICE_HANDLERS,
    DEFAULT_EOXS_OWS_EXCEPTION_HANDLERS
)
from eoxserver.core.util.importtools import import_string
from eoxserver.services.ows.decoders import get_decoder
from eoxserver.services.exceptions import (
    ServiceNotSupportedException, VersionNotSupportedException,
    VersionNegotiationException, OperationNotSupportedException,
    HTTPMethodNotAllowedError,
)
from eoxserver.services.ows.common.v20.exceptionhandler import (
    OWS20ExceptionHandler
)


logger = logging.getLogger(__name__)


ALLOWED_HTTP_METHODS = ["GET", "POST", "OPTIONS"]

SERVICE_HANDLERS = None
GET_SERVICE_HANDLERS = None
POST_SERVICE_HANDLERS = None
EXCEPTION_HANDLERS = None


def _setup_handlers():
    global SERVICE_HANDLERS
    global GET_SERVICE_HANDLERS
    global POST_SERVICE_HANDLERS
    global EXCEPTION_HANDLERS

    SERVICE_HANDLERS = [
        import_string(identifier)
        for identifier in getattr(
            settings, 'EOXS_SERVICE_HANDLERS', DEFAULT_EOXS_OWS_SERVICE_HANDLERS
        )
    ]

    GET_SERVICE_HANDLERS = [
        service_handler
        for service_handler in SERVICE_HANDLERS
        if 'GET' in service_handler.methods
    ]

    POST_SERVICE_HANDLERS = [
        service_handler
        for service_handler in SERVICE_HANDLERS
        if 'POST' in service_handler.methods
    ]

    EXCEPTION_HANDLERS = [
        import_string(identifier)
        for identifier in getattr(
            settings, 'EOXS_EXCEPTION_HANDLERS',
            DEFAULT_EOXS_OWS_EXCEPTION_HANDLERS
        )
    ]


class OptionsRequestHandler(object):
    """ Dummy request handler class to respond to HTTP OPTIONS requests.
    """
    def handle(self, request):

        def add_required_headers(headers, required_headers):
            """ Make sure the required headers are included in the list. """
            headers_lc = set(header.lower() for header in headers)
            for required_header in required_headers:
                if required_header.lower() not in headers_lc:
                    headers.append(required_header)
            return headers

        # return an empty 200 response
        response = HttpResponse()
        response["Access-Control-Allow-Methods"] = ", ".join(
            ALLOWED_HTTP_METHODS
        )
        headers = [
            header.strip() for header in
            request.META.get(
                "HTTP_ACCESS_CONTROL_REQUEST_HEADERS", ""
            ).split(",")
            if header
        ]
        headers = add_required_headers(headers, ['Content-Type'])
        response["Access-Control-Allow-Headers"] = ", ".join(headers)
        return response


def query_service_handler(request):
    """ Tries to find the correct service handler for a given request. The
    request ``method`` can either be "POST" (in which case the request body
    is parsed as XML) or "GET" (in which case the request is parsed
    as "KVP").

    If necessary a version negotiation is conducted, following OWS
    guidelines.

    :param request: a :class:`Django HttpRequest <django.http.HttpRequest>`
                    object
    :returns: the request handler component for the given request
    :raises ServiceNotSupportedException: if the service is not supported
                                          by any component
    :raises VersionNotSupportedException: if the specified version is not
                                          supported
    :raises OperationNotSupportedException: if the specified request
                                            operation is not supported
    """
    if SERVICE_HANDLERS is None:
        _setup_handlers()

    decoder = get_decoder(request)

    if request.method == "GET":
        handlers = GET_SERVICE_HANDLERS
    elif request.method == "POST":
        handlers = POST_SERVICE_HANDLERS
    elif request.method == "OPTIONS":
        return OptionsRequestHandler()
    else:
        raise HTTPMethodNotAllowedError(
            "The %s HTTP method is not allowed!" % request.method,
            ALLOWED_HTTP_METHODS
        )

    version = decoder.version
    if version is None:
        accepted_versions = decoder.acceptversions
        handlers = filter_handlers(
            handlers, decoder.service, accepted_versions, decoder.request
        )
        return version_negotiation(handlers, accepted_versions)()

    # check that the service is supported
    handlers = [
        handler
        for handler in handlers
        if handler_supports_service(handler, decoder.service)
    ]
    if not handlers:
        raise ServiceNotSupportedException(decoder.service)

    # check that the required version is enabled
    handlers_ = [
        handler for handler in handlers if decoder.version in handler.versions
    ]

    if not handlers_:
        # old style version negotiation shall always return capabilities
        if decoder.request == "GETCAPABILITIES":
            handlers = [sorted(
                filter(
                    lambda h: decoder.request == h.request.upper(), handlers
                ), key=lambda h: max(h.versions), reverse=True
            )[0]]
        else:
            raise VersionNotSupportedException(
                decoder.service, decoder.version
            )
    else:
        handlers = handlers_

    # check that the required operation is supported and sort by the highest
    # version supported in descending manner
    handlers = sorted(
        filter(
            lambda h: decoder.request == h.request.upper(), handlers
        ), key=lambda h: max(h.versions), reverse=True
    )

    if not handlers:
        operation = decoder.request
        raise OperationNotSupportedException(
            "Operation '%s' is not supported." % operation, operation
        )

    # return the handler with the highest version
    logger.debug("Handling '%s' request for '%s' service version '%s'." %
                 (handlers[0].request, handlers[0].service,
                  handlers[0].versions[0]))
    return handlers[0]()


def query_exception_handler(request):
    if EXCEPTION_HANDLERS is None:
        _setup_handlers()

    try:
        decoder = get_decoder(request)
        handlers = sorted([
                handler()
                for handler in EXCEPTION_HANDLERS
                if handler_supports_service(handler, decoder.service)
            ],
            key=lambda h: max(h.versions), reverse=True
        )

        # try to get the correctly versioned exception handler
        if decoder.version:
            for handler in handlers:
                if decoder.version in handler.versions:
                    return handler
        else:
            # return the exception handler with the highest version,
            # if one is available
            return handlers[0]
    except:
        # swallow any exception here, because we *really* need a handler
        # to correctly show the exception.
        pass

    # last resort fallback is a plain OWS exception handler
    return OWS20ExceptionHandler()


def version_negotiation(handlers, accepted_versions=None):
    version_to_handler = {}
    for handler in handlers:
        for version in handler.versions:
            version_to_handler.setdefault(version, handler)

    available_versions = sorted(version_to_handler.keys(), reverse=True)
    if not available_versions:
        raise VersionNegotiationException()

    if not accepted_versions:
        return version_to_handler[available_versions[0]]

    for accepted_version in accepted_versions:
        for available_version in available_versions:
            if accepted_version == available_version:
                return version_to_handler[available_version]

    raise VersionNegotiationException()


def filter_handlers(handlers=None, service=None, versions=None, request=None,
                    method=None):
    """ Utility function to filter the given OWS service handlers by their
        attributes 'service', 'versions' and 'request'.
    """
    if SERVICE_HANDLERS is None:
        _setup_handlers()

    handlers = handlers or SERVICE_HANDLERS

    service = service.upper() if service is not None else None
    request = request.upper() if request is not None else None

    if service:
        handlers = [
            handler
            for handler in handlers if handler_supports_service(handler, service)
        ]

    if request:
        handlers = [
            handler
            for handler in handlers
            if handler.request.upper() == request
        ]

    if versions:
        handlers = [
            handler for handler in handlers
            if any(version in handler.versions for version in versions)
        ]

    if method:
        handlers = [
            handler for handler in handlers
            if method in handler.methods
        ]

    return handlers


def handler_supports_service(handler, service=None):
    """ Convenience method to check whether or not a handler supports a service.
    """
    if isinstance(handler.service, string_types):
        return handler.service.upper() == service
    else:
        return service in handler.service



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2020 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import json
from os.path import isabs, join, dirname, normpath
from urllib.parse import urlparse

from django.contrib.gis.geos import GEOSGeometry
from django.db.models import Q
from django.db import transaction

from eoxserver.contrib import osr
from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.backends import models as backends
from eoxserver.backends.access import gdal_open
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.registration.exceptions import (
    RegistrationError
)
from eoxserver.resources.coverages.registration.product import create_metadata
from eoxserver.resources.coverages.registration.base import get_grid
from eoxserver.resources.coverages.metadata.coverage_formats import (
    get_reader_by_test
)


def get_product_type_name(stac_item):
    """ Create a ProductType name from a STAC Items metadata
    """

    properties = stac_item['properties']
    assets = stac_item['assets']

    parts = []

    platform = properties.get('platform') or properties.get('eo:platform')
    instruments = properties.get('instruments') or \
        properties.get('eo:instruments')
    constellation = properties.get('constellation') or \
        properties.get('eo:constellation')
    mission = properties.get('mission') or properties.get('eo:mission')

    if platform:
        parts.append(platform)

    if instruments:
        parts.extend(instruments)

    if constellation:
        parts.append(constellation)

    if mission:
        parts.append(mission)

    bands = properties.get('eo:bands')
    if not bands:
        bands = []
        for asset in assets.values():
            bands.extend(asset.get('eo:bands'), [])

    parts.extend([band['name'] for band in bands])

    if not parts:
        raise RegistrationError(
            'Failed to generate Product type name from metadata'
        )

    return '_'.join(parts)


@transaction.atomic
def register_stac_product(location, stac_item, product_type=None, storage=None,
                          replace=False):
    """ Registers a single parsed STAC item as a Product. The
        product type to be used can be specified via the product_type_name
        argument.
    """

    identifier = stac_item['id']
    replaced = False

    if replace:
        if models.Product.objects.filter(identifier=identifier).exists():
            models.Product.objects.filter(identifier=identifier).delete()
            replaced = True

    geometry = stac_item['geometry']
    properties = stac_item['properties']
    assets = stac_item['assets']

    # fetch the product type by name, metadata or passed object
    if isinstance(product_type, models.ProductType):
        pass
    if isinstance(product_type, str):
        product_type = models.ProductType.objects.get(name=product_type)
    else:
        product_type = models.ProductType.objects.get(
            name=get_product_type_name(stac_item)
        )

    if isinstance(storage, str):
        storage = backends.Storage.objects.get(name=storage)

    footprint = GEOSGeometry(json.dumps(geometry))
    if 'start_datetime' in properties and 'end_datetime' in properties:
        start_time = parse_iso8601(properties['start_datetime'])
        end_time = parse_iso8601(properties['end_datetime'])
    else:
        start_time = end_time = parse_iso8601(properties['datetime'])

    # check if the product already exists
    if models.Product.objects.filter(identifier=identifier).exists():
        if replace:
            models.Product.objects.filter(identifier=identifier).delete()
        else:
            raise RegistrationError('Product %s already exists' % identifier)

    product = models.Product.objects.create(
        identifier=identifier,
        begin_time=start_time,
        end_time=end_time,
        footprint=footprint,
        product_type=product_type,
    )

    metadata = {}
    simple_mappings = {
        'eo:cloud_cover': 'cloud_cover',
        'sar:instrument_mode': 'sensor_mode',
        'sat:relative_orbit': 'orbit_number',
        'view:incidence_angle': [
            'minimum_incidence_angle', 'maximum_incidence_angle'
        ],
        'view:sun_azimuth': 'illumination_azimuth_angle',
        'view:sun_elevation': 'illumination_elevation_angle',

    }

    for stac_key, field_name in simple_mappings.items():
        value = properties.get(stac_key)
        if value:
            if isinstance(field_name, str):
                metadata[field_name] = value
            else:
                for name in field_name:
                    metadata[name] = value

    # 'sar:frequency_band'
    # 'sar:center_frequency'
    # doppler_frequency ?
    # 'sar:product_type' #
    # 'sar:resolution_range'
    # 'sar:resolution_azimuth'
    # 'sar:pixel_spacing_range'
    # 'sar:pixel_spacing_azimuth'
    # 'sar:looks_range'
    # 'sar:looks_azimuth'
    # 'sar:looks_equivalent_number'
    # 'view:azimuth'

    complex_mappings = {
        'sar:polarizations': (
            'polarization_channels', lambda v: ', '.join(v)
        ),
        'sar:observation_direction': (
            'antenna_look_direction', lambda v: v.upper()
        ),
        'sat:orbit_state': (
            'orbit_direction', lambda v: v.upper()
        ),
    }

    for stac_key, field_desc in complex_mappings.items():
        raw_value = properties.get(stac_key)
        if raw_value:
            field_name, prep = field_desc
            value = prep(raw_value)
            if isinstance(field_name, str):
                metadata[field_name] = value
            else:
                for name in field_name:
                    metadata[name] = value

    # actually create the metadata object
    create_metadata(product, metadata)

    for asset_name, asset in assets.items():
        bands = asset.get('eo:bands')
        if not bands:
            continue

        if not isinstance(bands, list):
            bands = [bands]

        band_names = [band['name'] for band in bands]
        coverage_type = models.CoverageType.objects.get(
            Q(allowed_product_types=product_type),
            *[
                Q(field_types__identifier=band_name)
                for band_name in band_names
            ]
        )
        coverage_id = '%s_%s' % (identifier, asset_name)

        # create the storage item
        parsed = urlparse(asset['href'])

        if not isabs(parsed.path):
            path = normpath(join(dirname(location), parsed.path))
        else:
            path = parsed.path

        arraydata_item = models.ArrayDataItem(
            location=path,
            storage=storage,
            band_count=len(bands),
        )

        coverage_footprint = footprint
        if 'proj:geometry' in asset:
            coverage_footprint = GEOSGeometry(
                json.dumps(asset['proj:geometry'])
            )

        # get/create Grid
        grid_def = None
        size = None
        origin = None

        shape = asset.get('proj:shape') or properties.get('proj:shape')
        transform = asset.get('proj:transform') or \
            properties.get('proj:transform')
        epsg = asset.get('proj:epsg') or properties.get('proj:epsg')

        if shape:
            size = shape

        if transform:
            origin = [transform[transform[0], transform[3]]]

        if epsg and transform:
            sr = osr.SpatialReference(epsg)
            axis_names = ['x', 'y'] if sr.IsProjected() else ['long', 'lat']
            grid_def = {
                'coordinate_reference_system': epsg,
                'axis_names': axis_names,
                'axis_types': ['spatial', 'spatial'],
                'axis_offsets': [transform[1], transform[5]],
            }

        if not grid_def or not size or not origin:
            ds = gdal_open(arraydata_item)
            reader = get_reader_by_test(ds)
            if not reader:
                raise RegistrationError(
                    'Failed to get metadata reader for coverage'
                )
            values = reader.read(ds)
            grid_def = values['grid']
            size = values['size']
            origin = values['origin']

        grid = get_grid(grid_def)

        if models.Coverage.objects.filter(identifier=coverage_id).exists():
            if replace:
                models.Coverage.objects.filter(identifier=coverage_id).delete()
            else:
                raise RegistrationError(
                    'Coverage %s already exists' % coverage_id
                )

        coverage = models.Coverage.objects.create(
            identifier=coverage_id,
            footprint=coverage_footprint,
            begin_time=start_time,
            end_time=end_time,
            grid=grid,
            axis_1_origin=origin[0],
            axis_2_origin=origin[1],
            axis_1_size=size[0],
            axis_2_size=size[1],
            coverage_type=coverage_type,
            parent_product=product,
        )

        arraydata_item.coverage = coverage
        arraydata_item.full_clean()
        arraydata_item.save()

    # TODO: browses if possible

    return (product, replaced)


@transaction.atomic
def create_product_type_from_stac_collection(stac_collection,
                                             product_type_name=None):
    """
    """
    pass


@transaction.atomic
def create_product_type_from_stac_item(stac_item, product_type_name=None,
                                       ignore_existing=False):
    """ Creates a ProductType from a parsed STAC Item. Also creates all
        related CoverageTypes and their interned FieldTypes.

        Returns the ProductType and a boolean, indicating whether it was
        created or did already exist.
    """

    if product_type_name is None:
        product_type_name = get_product_type_name(stac_item)

    existing = models.ProductType.objects.filter(
        name=product_type_name
    ).first()
    if existing and ignore_existing:
        return (existing, False)
    elif existing:
        raise RegistrationError(
            'Product type %s already exists' % product_type_name
        )

    properties = stac_item['properties']
    assets = stac_item['assets']

    # list of Asset ID + bands
    bands_list = []

    for asset_name, asset in assets.items():
        if 'eo:bands' in asset:
            bands_list.append(
                (
                    asset_name,
                    asset['eo:bands']
                    if isinstance(asset['eo:bands'], list) else
                    [asset['eo:bands']]
                )
            )

    if not bands_list and 'eo:bands' in properties:
        bands_list.extend(
            (band['name'], [band])
            for band in properties['eo:bands']
        )

    if not bands_list:
        raise RegistrationError(
            'Failed to extract band defintion from STAC Item'
        )

    # create product type itself
    product_type = models.ProductType.objects.create(name=product_type_name)

    # iterate over bands and create coverage types
    for name, bands in bands_list:
        coverage_type = models.CoverageType.objects.create(
            name='%s_%s' % (product_type_name, name)
        )
        product_type.allowed_coverage_types.add(coverage_type)

        for i, band in enumerate(bands):
            models.FieldType.objects.create(
                coverage_type=coverage_type,
                index=i,
                identifier=band['name'],
                definition=band.get('common_name'),
            )

    browse_defs = properties.get('brow:browses', {})
    for browse_name, browse in browse_defs.items():
        browse_bands = browse.get('bands')
        range_ = browse.get('range')
        expression = browse.get('expression')
        browse_def = {'name': browse_name}
        if browse_bands:
            if len(browse_bands) not in (1, 3, 4):
                raise RegistrationError(
                    'Failed to create browse type: wrong number of bands'
                )
            browse_def['red_or_grey_expression'] = browse_bands[0]
            browse_def['red_or_grey_range_min'] = range_[0]
            browse_def['red_or_grey_range_max'] = range_[1]
            if len(browse_bands) >= 3:
                browse_def['green_expression'] = browse_bands[1]
                browse_def['green_range_min'] = range_[0]
                browse_def['green_range_max'] = range_[1]
                browse_def['blue_expression'] = browse_bands[2]
                browse_def['blue_range_min'] = range_[0]
                browse_def['blue_range_max'] = range_[1]
            if len(browse_bands) == 3:
                browse_def['alpha_expression'] = browse_bands[3]
                browse_def['alpha_range_min'] = range_[0]
                browse_def['alpha_range_max'] = range_[1]

        elif expression:
            browse_def['red_or_grey_expression'] = expression

        models.BrowseType.objects.create(
            product_type=product_type,
            **browse_def,
        )

    return (product_type, True)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class WCSCapabilitiesRendererInterface(object):
    """ Interface for WCS Capabilities renderers.
    """

    def render(self, params):
        """ Render the capabilities including information about the given 
            coverages.
        """

    def supports(self, params):
        """ Returns a boolean value to indicate whether or not the renderer is 
            able to render the capabilities with the given parameters.
        """


class WCSCoverageDescriptionRendererInterface(object):
    """ Interface for coverage description renderers.
    """

    def render(self, params):
        """ Render the description of the given coverages.
        """

    def supports(self, params):
        """ Returns a boolean value to indicate whether or not the renderer is 
            able to render the coverage and the given WCS version.
        """


class WCSCoverageRendererInterface(object):
    """ Interface for coverage renderers.
    """

    def render(self, params):
        """ Render the coverage with the given parameters.
        """

    def supports(self, params):
        """ Returns a boolean value to indicate whether or not the renderer is 
            able to render the coverage with the given parameters.
        """


class PackageWriterInterface(object):
    """ Interface for package writers.
    """

    def supports(self, format, params):
        """ Return a boolen value, whether or not a writer supports a given 
            format.
        """

    def create_package(self, filename, format, params):
        """ Create a package, which the encoder can later add items to with the 
            `cleanup` and `add_to_package` method.
        """

    def cleanup(self, package):
        """ Perform any necessary cleanups, like closing files, etc.
        """

    def add_to_package(self, package, file_obj, size, location):
        """ Add the file object to the package, that is returned by the 
            `create_package` method.
        """

    def get_mime_type(self, package, format, params):
        """ Retrieve the output mime type for the given package and/or format
            specifier.
        """

    def get_file_extension(self, package, format, params):
        """ Retrieve the file extension for the given package and format 
            specifier.
        """


class EncodingExtensionInterface(object):
    def supports(self, format, options):
        """ Return a boolen value, whether or not an encoding extension 
            supports a given format.
        """

    def parse_encoding_params(self, request):
        """ Return a dict, containing all additional encoding parameters from a 
            given request.
        """



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import sys
import json
import re

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction, IntegrityError

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage coverage types. This command uses sub-commands for the
        specific tasks: create, delete
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create',
            help='Create a new coverage type.'
        )
        import_parser = self.add_subparser(parser, 'import')
        delete_parser = self.add_subparser(parser, 'delete',
            help='Delete a coverage type.'
        )
        list_parser = self.add_subparser(parser, 'list')

        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'name', nargs=1, help='The coverage type name. Mandatory.'
            )

        create_parser.add_argument(
            '--field-type', action='append', nargs=5,
            metavar=(
                'identifier', 'description', 'definition', 'unit-of-measure',
                'wavelength'
            ),
            dest='field_types', default=[],
            help=(
                'Add a field type to the coverage type.'
            )
        )

        import_parser.add_argument(
            'locations', nargs='*',
            help='The location(s) of the coverage type schema(s). Mandatory.'
        )
        import_parser.add_argument(
            '--in, -i', dest='stdin', action="store_true", default=False,
            help='Read the definition from stdin instead from a file.'
        )

        delete_parser.add_argument(
            '--force', '-f', action='store_true', default=False,
            help='Also remove all collections associated with that type.'
        )

        list_parser.add_argument(
            '--no-detail', action="store_false", default=True, dest='detail',
            help="Disable the printing of details of the coverage type."
        )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, insert and exclude.
        """
        if subcommand == "create":
            self.handle_create(kwargs.pop('name')[0], *args, **kwargs)
        elif subcommand == "import":
            self.handle_import(*args, **kwargs)
        elif subcommand == "export":
            self.handle_export(*args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(kwargs.pop('name')[0], *args, **kwargs)
        elif subcommand == "list":
            self.handle_list(*args, **kwargs)

    def handle_create(self, name, field_types, **kwargs):
        """ Handle the creation of a new coverage type.
        """
        coverage_type = self._create_coverage_type(name)

        self._create_field_types(coverage_type, {}, [
            dict(
                identifier=field_type_definition[0],
                description=field_type_definition[1],
                definition=field_type_definition[2],
                unit_of_measure=field_type_definition[3],
                wavelength=field_type_definition[4]
            )
            for field_type_definition in field_types
        ])

        print('Successfully created coverage type %r' % name)

    def handle_import(self, locations, *args, **kwargs):
        def _import(definitions):
            if isinstance(definitions, dict):
                definitions = [definitions]

            for definition in definitions:
                self._import_definition(definition)

        if kwargs['stdin']:
            try:
                _import(json.load(sys.stdin))
            except ValueError:
                raise CommandError('Could not parse JSON from stdin')
        else:
            for location in locations:
                with open(location) as f:
                    try:
                        _import(json.load(f))
                    except ValueError:
                        raise CommandError(
                            'Could not parse JSON from %r' % location
                        )

    def handle_export(self, name, *args, **kwargs):
        pass

    def handle_delete(self, name, force, **kwargs):
        """ Handle the deletion of a collection type
        """
        try:
            coverage_type = models.CoverageType.objects.get(name=name)

            if force:
                coverages = models.Coverage.objects.filter(
                    coverage_type=coverage_type
                )
                coverages.delete()
            nil_values = set(models.NilValue.objects.filter(
                field_types__coverage_type=coverage_type
            ))

            coverage_type.delete()

            # delete orphaned nil-values
            for nil_value in nil_values:
                if nil_value.field_types.count() == 0:
                    nil_value.delete()

        except models.CoverageType.DoesNotExist:
            raise CommandError('No such coverage type: %r' % name)

        print('Successfully deleted coverage type %r' % name)

    def handle_list(self, detail, *args, **kwargs):
        """ Handle the listing of product types
        """
        for coverage_type in models.CoverageType.objects.all():
            print(coverage_type.name)
            if detail:
                for coverage_type in coverage_type.field_types.all():
                    print("\t%s" % coverage_type.identifier)

    def _import_definition(self, definition):
        name = str(definition['name'])
        coverage_type = self._create_coverage_type(name)
        field_type_definitions = (
            definition.get('field_type') or definition.get('bands')
        )
        self._create_field_types(
            coverage_type, definition, field_type_definitions
        )
        self.print_msg('Successfully imported coverage type %r' % name)

    def _create_coverage_type(self, name):
        try:
            return models.CoverageType.objects.create(name=name)
        except IntegrityError:
            raise CommandError("Coverage type %r already exists." % name)

    def _create_field_types(self, coverage_type, coverage_type_definition,
                            field_type_definitions):
        for i, field_type_definition in enumerate(field_type_definitions):
            uom = (
                field_type_definition.get('unit_of_measure') or
                field_type_definition.get('uom')
            )

            field_type = models.FieldType(
                coverage_type=coverage_type,
                index=i,
                identifier=field_type_definition.get('identifier'),
                description=field_type_definition.get('description'),
                definition=field_type_definition.get('definition'),
                unit_of_measure=uom,
                wavelength=field_type_definition.get('wavelength'),
                significant_figures=field_type_definition.get(
                    'significant_figures'
                )
            )

            if 'numbits' in field_type_definition:
                field_type.numbits = field_type_definition['numbits']
            if 'signed' in field_type_definition:
                field_type.signed = field_type_definition['signed']
            if 'is_float' in field_type_definition:
                field_type.is_float = field_type_definition['is_float']

            # per field data type
            if 'data_type' in field_type_definition:
                field_type.numbits, field_type.signed, field_type.is_float = \
                    self._parse_data_type(field_type_definition['data_type'])

            # global data type
            elif 'data_type' in coverage_type_definition:
                field_type.numbits, field_type.signed, field_type.is_float = \
                    self._parse_data_type(coverage_type_definition['data_type'])

            field_type.full_clean()
            field_type.save()

            nil_value_definitions = field_type_definition.get('nil_values', [])
            for nil_value_definition in nil_value_definitions:
                # TODO: in Django 1.11 a `get` query did not work
                nil_value = next(iter(models.NilValue.objects.filter(
                    value=nil_value_definition['value'],
                    reason=nil_value_definition['reason'],
                    field_types__coverage_type=coverage_type,
                )), None)
                if not nil_value:
                    nil_value = models.NilValue.objects.create(
                        value=nil_value_definition['value'],
                        reason=nil_value_definition['reason'],
                    )

                nil_value.field_types.add(field_type)
                nil_value.save()

            allowed_value_ranges = field_type_definition.get(
                'allowed_value_ranges', []
            )
            for allowed_value_range_definition in allowed_value_ranges:
                models.AllowedValueRange.objects.create(
                    field_type=field_type,
                    start=allowed_value_range_definition[0],
                    end=allowed_value_range_definition[1]
                )

    def _parse_data_type(self, data_type):
        data_type = data_type.lower()
        is_float = data_type.startswith('float')
        signed = data_type.startswith('float') or data_type.startswith('int')
        try:
            if data_type == 'byte':
                numbits = 8
            else:
                numbits = int(
                    re.search(r'[a-zA-Z]+(\d*)', data_type).groups()[0]
                )
        except ValueError:
            numbits = None
        except AttributeError:
            raise CommandError('Invalid data type description %r' % data_type)
        return numbits, signed, is_float



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from itertools import chain

from lxml.etree import CDATA
from lxml.builder import ElementMaker
try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse
from django.utils.http import urlencode

from eoxserver.core.util.xmltools import etree, NameSpace, NameSpaceMap, typemap
from eoxserver.core.util.timetools import isoformat
from eoxserver.services.opensearch.formats.base import (
    BaseFeedResultFormat, ns_opensearch, ns_dc, ns_atom, ns_media, ns_owc
)


# namespace declarations
ns_georss = NameSpace("http://www.georss.org/georss", "georss")
ns_gml = NameSpace("http://www.opengis.net/gml", "gml")

# namespace map
nsmap = NameSpaceMap(
    ns_georss, ns_gml, ns_opensearch, ns_dc, ns_atom, ns_media, ns_owc
)

# Element factories
GEORSS = ElementMaker(namespace=ns_georss.uri, nsmap=nsmap)
GML = ElementMaker(namespace=ns_gml.uri, nsmap=nsmap)
RSS = ElementMaker(typemap=typemap)


class RSSResultFormat(BaseFeedResultFormat):
    """ RSS result format.
    """

    mimetype = "application/rss+xml"
    name = "rss"

    def encode(self, request, collection_id, queryset, search_context):
        # prepare RSS factory with additional namespaces from search context
        namespaces = dict(nsmap)
        namespaces.update(search_context.namespaces)
        RSS = ElementMaker(namespace=None, nsmap=namespaces)

        tree = RSS("rss",
            RSS("channel",
                RSS("title", "%s Search" % collection_id),
                RSS("link", request.build_absolute_uri()),
                RSS("description"),
                *chain(
                    self.encode_opensearch_elements(search_context),
                    self.encode_feed_links(request, search_context), [
                        self.encode_item(request, item, search_context)
                        for item in queryset
                    ]
                )
            ),
            version="2.0"
        )
        return etree.tostring(tree, pretty_print=True)

    def encode_item(self, request, item, search_context):
        link_url = request.build_absolute_uri(
            "%s?%s" % (
                reverse("ows"), urlencode(dict(
                    service="WCS",
                    version="2.0.1",
                    request="DescribeCoverage",
                    coverageId=item.identifier,
                ))
            )
        )

        rss_item = RSS("item",
            RSS("title", item.identifier),
            RSS("description", CDATA(item.identifier)),
            RSS("link", link_url),
        )

        if "geo" in search_context.parameters:
            rss_item.append(RSS("guid", request.build_absolute_uri()))
        else:
            rss_item.append(RSS("guid", item.identifier, isPermaLink="false"))

        rss_item.extend(self.encode_item_links(request, item))

        # TODO: remove this for the general dc:date?
        if item.begin_time and item.end_time:
            rss_item.append(
                GML("TimePeriod",
                    GML("beginPosition", isoformat(item.begin_time)),
                    GML("endPosition", isoformat(item.end_time)),
                    **{ns_gml("id"): item.identifier}
                )
            )

        rss_item.extend(self.encode_spatio_temporal(item))
        return rss_item



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
This module provides an implementation of a system configuration that relies
on different configuration files.
"""

import imp
from os.path import join, getmtime
from sys import prefix
import threading
import logging
from time import time

try:
    from ConfigParser import RawConfigParser
except ImportError:
    from configparser import RawConfigParser


from django.conf import settings


config_lock = threading.RLock()
logger = logging.getLogger(__name__)

# configuration singleton
_cached_config = None
_last_access_time = None


def get_eoxserver_config():
    """ Returns the EOxServer config as a :class:`ConfigParser.RawConfigParser`
    """
    with config_lock:
        if not _cached_config or \
                getmtime(get_instance_config_path()) > _last_access_time:
            reload_eoxserver_config()

    return _cached_config


def reload_eoxserver_config():
    """ Triggers the loading or reloading of the EOxServer config as a
        :class:`ConfigParser.RawConfigParser`.
    """
    global _cached_config, _last_access_time
    _, eoxs_path, _ = imp.find_module("eoxserver")
    paths = [
        join(eoxs_path, "conf", "default.conf"),
        join(prefix, "eoxserver/conf/default.conf"),
        get_instance_config_path()
    ]

    logger.info(
        "%soading the EOxServer configuration. Using paths: %s."
        % ("Rel" if _cached_config else "L", ", ".join(paths))
    )

    with config_lock:
        _cached_config = RawConfigParser()
        _cached_config.read(paths)
        _last_access_time = time()


def get_instance_config_path():
    """ Convenience function to get the path to the instance config.
    """
    return join(settings.PROJECT_DIR, "conf", "eoxserver.conf")



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage mask types. This command uses sub-commands for the
        specific tasks: create, delete, list
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        list_parser = self.add_subparser(parser, 'list')

        for parser in [create_parser, delete_parser, list_parser]:
            parser.add_argument(
                'product_type_name', nargs=1,
                help='The product type name. Mandatory.'
            )

        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'mask_type_name', nargs=1,
                help='The mask type name. Mandatory.'
            )

        create_parser.add_argument(
            '--validity',
            dest='validity', action='store_true', default=False,
            help='Whether this mask is a validity mask or an invalidity mask. '
                 'Defaults to invalidity.'
        )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: create, delete.
        """
        if subcommand == "create":
            self.handle_create(
                kwargs.pop('product_type_name')[0],
                kwargs.pop('mask_type_name')[0],
                *args, **kwargs
            )
        elif subcommand == "delete":
            self.handle_delete(
                kwargs.pop('product_type_name')[0],
                kwargs.pop('mask_type_name')[0],
                *args, **kwargs
            )
        elif subcommand == "list":
            self.handle_list(
                kwargs.pop('product_type_name')[0], *args, **kwargs
            )

    def handle_create(self, product_type_name, mask_type_name, validity,
                      *args, **kwargs):
        """ Handle the creation of a new mask type.
        """

        try:
            product_type = models.ProductType.objects.get(name=product_type_name)
        except models.ProductType.DoesNotExist:
            raise CommandError(
                'Product type %r does not exist' % product_type_name
            )

        models.MaskType.objects.create(
            product_type=product_type,
            name=mask_type_name,
            validity=validity,
        )

        print(
            'Successfully created mask type %r for product type %r'
            % (mask_type_name, product_type_name)
        )

    def handle_delete(self, product_type_name, mask_type_name, **kwargs):
        """ Handle the deletion of a mask type
        """

        try:
            product_type = models.ProductType.objects.get(name=product_type_name)
        except models.ProductType.DoesNotExist:
            raise CommandError('No such product type %r' % product_type_name)

        mask_type = product_type.mask_types.get(name=mask_type_name)

        mask_type.delete()

        print(
            'Successfully deleted mask type %r for product type %r'
            % (mask_type_name, product_type_name)
        )

    def handle_list(self, product_type_name, *args, **kwargs):
        """ Handle the listing of mask types
        """
        try:
            product_type = models.ProductType.objects.get(name=product_type_name)
        except models.ProductType.DoesNotExist:
            raise CommandError('No such product type %r' % product_type_name)

        for mask_type in product_type.mask_types.all():
            print("%s %s" % (mask_type.name, '(validity)' if mask_type.validity else ''))



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from lxml.builder import ElementMaker

from eoxserver.core.util.xmltools import XMLEncoder, NameSpace, NameSpaceMap

ns_wms = NameSpace("http://www.opengis.net/wms")
ns_xlink = NameSpace("http://www.w3.org/1999/xlink", "xlink")
nsmap = NameSpaceMap(ns_wms, ns_xlink)

WMS = ElementMaker(namespace=ns_wms.uri, nsmap=nsmap)


class WMS13Encoder(XMLEncoder):
    def encode_capabilities(self, config, ows_url, srss, formats, info_formats,
                            layer_descriptions):
        return WMS("WMS_Capabilities",
            WMS("Service",
                WMS("Name", config.name),
                WMS("Title", config.title),
                WMS("Abstract", config.abstract),
                WMS("KeywordList", *[
                    WMS("Keyword", keyword)
                    for keyword in config.keywords
                ]),
                WMS("OnlineResource", config.onlineresource),

                WMS("ContactInformation",
                    WMS("ContactPersonPrimary",
                        WMS("ContactPerson", config.individual_name),
                        WMS("ContactOrganization",  config.provider_name),
                    ),
                    WMS("ContactPosition", config.position_name),
                    WMS("ContactAddress",
                        WMS("AddressType", "postal"),
                        WMS("Address", config.delivery_point),
                        WMS("City", config.city),
                        WMS("StateOrProvince", config.administrative_area),
                        WMS("PostCode", config.postal_code),
                        WMS("Country", config.country),
                    ),
                    WMS("ContactVoiceTelephone", config.phone_voice),
                    WMS("ContactFacsimileTelephone", config.phone_facsimile),
                    WMS("ContactElectronicMailAddress",
                        config.electronic_mail_address
                    ),
                ),
                WMS("Fees", config.fees),
                WMS("AccessConstraints", config.access_constraints),

                # TODO:
                # <LayerLimit>16</LayerLimit>
                # <MaxWidth>2048</MaxWidth>
                # <MaxHeight>2048</MaxHeight>
            ),
            WMS("Capability",
                WMS("Request",
                    WMS("GetCapabilities",
                        WMS("Format", "text/xml"),
                        self.encode_dcptype(ows_url)
                    ),
                    WMS("GetMap", *[
                            WMS("Format", frmt.mimeType)
                            for frmt in formats
                        ] + [
                            self.encode_dcptype(ows_url)
                        ]
                    ),
                    WMS("GetFeatureInfo",
                        WMS("Format",
                            # TODO
                        ),
                        self.encode_dcptype(ows_url)
                    ),
                    # TODO: describe layer?
                ),
                WMS("Exception",
                    WMS("Format", "XML"),
                    WMS("Format", "INIMAGE"),
                    WMS("Format", "BLANK"),
                ),
                WMS("Layer",
                    WMS("Title", config.title),
                    *([
                        WMS("CRS", srs)
                        for srs in srss
                    ] + [
                        self.encode_bbox(
                            minx="-180", miny="-90", maxx="180", maxy="90"
                        )
                    ] + [
                        self.encode_layer(layer_description)
                        for layer_description in layer_descriptions
                    ])
                )
            ),
            version="1.3.0", updateSequence=config.update_sequence
        )

    def encode_dcptype(self, ows_url):
        return WMS("DCPType",
            WMS("HTTP",
                WMS("Get",
                    WMS("OnlineResource", **{
                        ns_xlink("href"): ows_url,
                        ns_xlink("type"): "simple"
                    })
                )
            )
        )

    def encode_bbox(self, minx, miny, maxx, maxy):
        return WMS("EX_GeographicBoundingBox",
            WMS("westBoundLongitude", minx),
            WMS("eastBoundLongitude", maxx),
            WMS("southBoundLatitude", miny),
            WMS("northBoundLatitude", maxy),
        )

    def encode_layer(self, layer_description):
        elems = [
            WMS("Name", layer_description.name)
        ]

        title = getattr(layer_description, 'title')
        if title:
            elems.append(WMS("Title", title))

        if layer_description.bbox:
            bbox = list(map(str, layer_description.bbox))
            elems.append(
                self.encode_bbox(
                    minx=bbox[0], miny=bbox[1], maxx=bbox[2], maxy=bbox[3]
                )
            )

        elems.extend(
            WMS("Style",
                WMS("Name", style),
                WMS("Abstract", style),
            ) for style in layer_description.styles
        )

        elems.extend(
            self.encode_layer(sub_layer)
            for sub_layer in layer_description.sub_layers
        )

        dimensions = []

        for dimension_name, dimension in layer_description.dimensions.items():
            if "min" in dimension and "max" in dimension and "step" in dimension:
                extent_text = "%s/%s/%s" % (
                    dimension["min"], dimension["max"], dimension["step"]
                )
            elif "values" in dimension:
                extent_text = ",".join(dimension["values"])

            dimension_elem = WMS("Dimension", extent_text, name=dimension_name)
            if "units" in dimension:
                dimension_elem.attrib["units"] = dimension["units"]

            if "default" in dimension:
                dimension_elem.attrib["default"] = dimension["default"]

            if dimension.get("multivalue"):
                dimension_elem.attrib["multipleValues"] = "1"

            dimensions.append(dimension_elem)

        elems.extend(dimensions)

        return WMS("Layer",
            *elems,
            queryable="1" if layer_description.queryable else "0"
        )



#-------------------------------------------------------------------------------
#
#  Library of primitive data type classes.
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from datetime import datetime, date, time, timedelta
from django.utils.dateparse import parse_date, parse_datetime, parse_time, utc
from django.utils.six import PY2, PY3, string_types
from django.utils.encoding import smart_text
from eoxserver.core.util.timetools import parse_duration

try:
    from datetime import timezone

    # as this class will be deprecated in Django 3.1, offer a constructor
    def FixedOffset(offset, name=None):
        if isinstance(offset, timedelta):
            pass
        else:
            offset = timedelta(minutes=offset)
        return timezone(offset) if name is None else timezone(offset, name)

except ImportError:
    from django.utils.timezone import FixedOffset


class BaseType(object):
    """ Base literal data type class.
        This class defines the class interface.
    """
    name = None  # to be replaced by a name
    if PY2:
        dtype = str
    elif PY3:
        dtype = bytes
    zero = None  # when applicable to be replaced by a proper zero value
    comparable = True  # indicate whether the type can be compared (<,>,==)

    @classmethod
    def parse(cls, raw_value):
        """ Cast or parse input to its proper representation."""
        return cls.dtype(raw_value)

    @classmethod
    def encode(cls, value):
        """ Encode value to a Unicode string."""
        return smart_text(value)

    @classmethod
    def get_diff_dtype(cls):  # difference type - change if differs from the base
        """ Get type of the difference of this type.
            E.g., `timedelta` for a `datetime`.
        """
        return cls

    @classmethod
    def as_number(cls, value):  # pylint: disable=unused-argument
        """ convert to a number (e.g., duration)"""
        raise TypeError("Data type %s cannot be converted to a number!" % cls)

    @classmethod
    def sub(cls, value0, value1):  # pylint: disable=unused-argument
        """ subtract value0 - value1 """
        raise TypeError("Data type %s cannot be subtracted!" % cls)


class Boolean(BaseType):
    """ Boolean literal data type class. """
    name = "boolean"
    dtype = bool

    @classmethod
    def parse(cls, raw_value):

        if isinstance(raw_value, string_types):
            raw_value = smart_text(raw_value.lower())
            if raw_value in ('1', 'true'):
                return True
            elif raw_value in ('0', 'false'):
                return False
            else:
                raise ValueError("Cannot parse boolean value '%s'!" % raw_value)
        else:
            return bool(raw_value)

    @classmethod
    def encode(cls, value):
        return u'true' if value else u'false'

    @classmethod
    def as_number(cls, value):
        return int(value)

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        return value0 - value1


class Integer(BaseType):
    """ Integer literal data type class. """
    name = "integer"
    dtype = int
    zero = 0

    @classmethod
    def encode(cls, value):
        """ Encode value to a Unicode string."""
        return smart_text(int(value))

    @classmethod
    def as_number(cls, value):
        return value

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        return value0 - value1


class Double(BaseType):
    """ Double precision float literal data type class. """
    name = "double"
    dtype = float
    zero = 0.0

    @classmethod
    def encode(cls, value):
        return u"%.15g" % cls.dtype(value)

    @classmethod
    def as_number(cls, value):
        return value

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        return value0 - value1


class String(BaseType):
    """ Unicode character string literal data type class. """
    name = "string"
    if PY2:
        dtype = unicode
    elif PY3:
        dtype = str
    encoding = 'utf-8'
    comparable = False  # disabled although Python implements comparable strings

    @classmethod
    def encode(cls, value):
        """ Encode value to a Unicode string."""
        try:
            return smart_text(value)
        except UnicodeDecodeError:
            return smart_text(value, cls.encoding)

    @classmethod
    def parse(cls, raw_value):
        return cls.dtype(raw_value)

    @classmethod
    def get_diff_dtype(cls):  # string has no difference
        return None


class Duration(BaseType):
    """ Duration (`datetime.timedelta`) literal data type class. """
    name = "duration"
    dtype = timedelta
    zero = timedelta(0)

    @classmethod
    def parse(cls, raw_value):
        if isinstance(raw_value, cls.dtype):
            return raw_value
        return parse_duration(raw_value)

    @classmethod
    def encode(cls, value):
        # NOTE: USE OF MONTH AND YEAR IS AMBIGUOUS! WE DO NOT ENCODE THEM!
        if not isinstance(value, cls.dtype):
            raise ValueError("Invalid value type '%s'!" % type(value))
        items = []
        if value.days < 0:
            items.append('-')
            value = -value
        items.append('P')
        if value.days != 0:
            items.append('%dD' % value.days)
        elif value.seconds == 0 and value.microseconds == 0:
            items.append('T0S')  # zero interval
        if value.seconds != 0 or value.microseconds != 0:
            minutes, seconds = divmod(value.seconds, 60)
            hours, minutes = divmod(minutes, 60)
            items.append('T')
            if hours != 0:
                items.append('%dH' % hours)
            if minutes != 0:
                items.append('%dM' % minutes)
            if value.microseconds != 0:
                items.append("%.6fS" % (seconds + 1e-6*value.microseconds))
            elif seconds != 0:
                items.append('%dS' % seconds)

        return smart_text("".join(items))

    @classmethod
    def as_number(cls, value):
        return 86400.0*value.days + 1.0*value.seconds + 1e-6*value.microseconds

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        return value0 - value1


class Date(BaseType):
    """ Date (`datetime.date`) literal data type class. """
    name = "date"
    dtype = date

    @classmethod
    def get_diff_dtype(cls):
        return Duration

    @classmethod
    def parse(cls, raw_value):
        if isinstance(raw_value, cls.dtype):
            return raw_value
        value = parse_date(raw_value)
        if value:
            return value
        raise ValueError("Could not parse ISO date from '%s'." % raw_value)

    @classmethod
    def encode(cls, value):
        if isinstance(value, cls.dtype):
            return smart_text(value.isoformat())
        raise ValueError("Invalid value type '%s'!" % type(value))

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        return value0 - value1


class Time(BaseType):
    """ Time (`datetime.time`) literal data type class. """
    name = "time"
    dtype = time
    # TODO: implement proper Time time-zone handling

    @classmethod
    def get_diff_dtype(cls):
        return Duration

    @classmethod
    def parse(cls, raw_value):
        if isinstance(raw_value, cls.dtype):
            return raw_value
        value = parse_time(raw_value)
        if value is not None:
            return value
        raise ValueError("Could not parse ISO time from '%s'." % raw_value)

    @classmethod
    def encode(cls, value):
        if isinstance(value, cls.dtype):
            return smart_text(value.isoformat())
        raise ValueError("Invalid value type '%s'!" % type(value))

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        aux_date = datetime.now().date()
        dt0 = datetime.combine(aux_date, value0)
        dt1 = datetime.combine(aux_date, value1)
        return dt0 - dt1


class DateTime(BaseType):
    """ Date-time (`datetime.datetime`) literal data type class. """
    name = "dateTime"
    dtype = datetime

    # tzinfo helpers
    UTC = utc               # Zulu-time TZ instance
    TZOffset = FixedOffset  # fixed TZ offset class, set minutes to instantiate

    @classmethod
    def get_diff_dtype(cls):
        return Duration

    @classmethod
    def parse(cls, raw_value):
        if isinstance(raw_value, cls.dtype):
            return raw_value
        value = parse_datetime(raw_value)
        if value:
            return value
        raise ValueError("Could not parse ISO date-time from '%s'." % raw_value)

    @classmethod
    def encode(cls, value):
        if isinstance(value, cls.dtype):
            return smart_text(cls._isoformat(value))
        raise ValueError("Invalid value type '%s'!" % type(value))

    @classmethod
    def sub(cls, value0, value1):
        """ subtract value0 - value1 """
        return value0 - value1

    @staticmethod
    def _isoformat(value):
        """ Covert date-time object to ISO 8601 date-time string. """
        if value.tzinfo and not value.utcoffset():
            return value.replace(tzinfo=None).isoformat("T") + "Z"
        return value.isoformat("T")


class DateTimeTZAware(DateTime):
    """ Time-zone aware date-time (`datetime.datetime`) literal data type class.

    This data-type is a variant of the `DateTime` which assures that
    the parsed date-time is time-zone aware and optionally
    also converted to a common target time-zone.

    The default time-zone applied to the unaware time-input is passed trough
    the constructor. By default the UTC time-zone is used.
    By default the target time-zone is set to None which means that
    the original time-zone is preserved.

    Unlike the `DateTime` this class must be instantiated and it cannot be used
    directly as a data-type.

    Constructor parameters:
        default_tz  default time-zone
        target_tz   optional target time-zone
    """
    def __init__(self, default_tz=DateTime.UTC, target_tz=None):
        self.default_tz = default_tz
        self.target_tz = target_tz

    def set_time_zone(self, value):
        """ Make a date-time value time-zone aware by setting the default
        time-zone and convert the time-zone if the target time-zone is given.
        """
        if value.tzinfo is None:
            value = value.replace(tzinfo=self.default_tz)
        if self.target_tz:
            value = value.astimezone(self.target_tz)
        return value

    def parse(self, raw_value):
        return self.set_time_zone(super(DateTimeTZAware, self).parse(raw_value))

    def encode(self, value):
        return super(DateTimeTZAware, self).encode(self.set_time_zone(value))


# mapping of plain Python types to data type classes
if PY3:

    DTYPES = {
        bytes: String,
        str: String,
        bool: Boolean,
        int: Integer,
        float: Double,
        date: Date,
        datetime: DateTime,
        time: Time,
        timedelta: Duration,
    }
elif PY2:

    DTYPES = {
        str: String,
        unicode: String,
        bool: Boolean,
        int: Integer,
        long: Integer,
        float: Double,
        date: Date,
        datetime: DateTime,
        time: Time,
        timedelta: Duration,
    }



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging
import operator

from django.contrib.gis.geos import Polygon, LineString

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config, enum
from eoxserver.contrib.osr import SpatialReference
from eoxserver.resources.coverages import crss
from eoxserver.services.exceptions import (
    InvalidAxisLabelException, InvalidSubsettingException,
    InvalidSubsettingCrsException
)


__all__ = ["Subsets", "Trim", "Slice"]

logger = logging.getLogger(__name__)


class Subsets(list):
    """ Convenience class to handle a variety of spatial and/or temporal
        subsets.

        :param iterable: an iterable of objects inheriting from :class:`Trim`
                         or :class:`Slice`
        :param crs: the CRS definition
        :param allowed_types: the allowed subset types. defaults to both
                              :class:`Trim` and :class:`Slice`
    """

    def __init__(self, iterable, crs=None, allowed_types=None):
        """ Constructor. Allows to add set the initial subsets
        """
        self.allowed_types = allowed_types if allowed_types is not None else (
            Trim, Slice
        )
        # Do a manual insertion here to assure integrity
        for subset in iterable:
            self.append(subset)

        self._crs = crs

    # List API

    def extend(self, iterable):
        """ See :meth:`list.extend` """
        for subset in iterable:
            self._check_subset(subset)
            super(Subsets, self).append(subset)

    def append(self, subset):
        """ See :meth:`list.append` """
        self._check_subset(subset)
        super(Subsets, self).append(subset)

    def insert(self, i, subset):
        """ See :meth:`list.insert` """
        self._check_subset(subset)
        super(Subsets, self).insert(i, subset)

    # Subset related stuff

    @property
    def has_x(self):
        """ Check if a subset along the X-axis is given. """
        return any(map(lambda s: s.is_x, self))

    @property
    def has_y(self):
        """ Check if a subset along the Y-axis is given. """
        return any(map(lambda s: s.is_y, self))

    @property
    def has_t(self):
        """ Check if a subset along the temporal axis is given. """
        return any(map(lambda s: s.is_temporal, self))

    @property
    def crs(self):
        """ Return the subset CRS definiton. """
        return self._crs

    @crs.setter
    def crs(self, value):
        """ Set the subset CRS definiton. """
        self._crs = value

    @property
    def srid(self):
        """ Tries to find the correct integer SRID for the crs.
        """
        crs = self.crs
        if crs is not None:
            srid = crss.parseEPSGCode(crs,
                (crss.fromURL, crss.fromURN, crss.fromShortCode)
            )
            if srid is None and not crss.is_image_crs(crs):
                raise InvalidSubsettingCrsException(
                    "Could not parse EPSG code from URI '%s'" % crs
                )
            return srid
        return None

    def get_filters(self, containment="overlaps"):
        """ Filter a :class:`Django QuerySet <django.db.models.query.QuerySet>`
        of objects inheriting from :class:`EOObject
        <eoxserver.resources.coverages.models.EOObject>`.

        :param queryset: the ``QuerySet`` to filter
        :param containment: either "overlaps" or "contains"
        :returns: a ``dict`` with the filters
        """
        filters = {}
        if not len(self):
            return filters

        bbox = [None, None, None, None]
        srid = self.srid

        if srid is None:
            srid = 4326
        max_extent = crss.crs_bounds(srid)
        tolerance = crss.crs_tolerance(srid)

        # check if time intervals are configured as "open" or "closed"
        config = get_eoxserver_config()
        reader = SubsetConfigReader(config)
        if reader.time_interval_interpretation == "closed":
            gt_op = "__gte"
            lt_op = "__lte"
        else:
            gt_op = "__gt"
            lt_op = "__lt"

        for subset in self:
            if isinstance(subset, Slice):
                is_slice = True
                value = subset.value
            elif isinstance(subset, Trim):
                is_slice = False
                low = subset.low
                high = subset.high
                # we need the value in case low == high
                value = low

            if subset.is_temporal:
                if is_slice or (high == low and containment == "overlaps"):
                    filters['begin_time__lte'] = value
                    filters['end_time__gte'] = value

                elif high == low:
                    filters['begin_time__gte'] = value
                    filters['end_time__lte'] = value

                else:
                    # check if the temporal bounds must be strictly contained
                    if containment == "contains":
                        if high is not None:
                            filters['end_time' + lt_op] = high
                        if low is not None:
                            filters['begin_time' + gt_op] = low
                    # or just overlapping
                    else:
                        if high is not None:
                            filters['begin_time' + lt_op] = high
                        if low is not None:
                            filters['end_time' + gt_op] = low

            else:
                if is_slice:
                    if subset.is_x:
                        line = LineString(
                            (value, max_extent[1]),
                            (value, max_extent[3])
                        )
                    else:
                        line = LineString(
                            (max_extent[0], value),
                            (max_extent[2], value)
                        )
                    line.srid = srid
                    if srid != 4326:
                        line.transform(4326)
                    filters['footprint__intersects'] = line

                else:
                    if subset.is_x:
                        bbox[0] = subset.low
                        bbox[2] = subset.high
                    else:
                        bbox[1] = subset.low
                        bbox[3] = subset.high

        if bbox != [None, None, None, None]:
            bbox = list(map(
                lambda v: v[0] if v[0] is not None else v[1],
                zip(bbox, max_extent)
            ))

            bbox[0] -= tolerance
            bbox[1] -= tolerance
            bbox[2] += tolerance
            bbox[3] += tolerance

            logger.debug(
                "Applying BBox %s with containment '%s'." % (bbox, containment)
            )

            poly = Polygon.from_bbox(bbox)
            poly.srid = srid

            if srid != 4326:
                poly.transform(4326)
            if containment == "overlaps":
                filters['footprint__intersects'] = poly
            elif containment == "contains":
                filters['footprint__within'] = poly

        return filters

    def filter(self, queryset, containment="overlaps"):
        """ Filter a :class:`Django QuerySet <django.db.models.query.QuerySet>`
        of objects inheriting from :class:`EOObject
        <eoxserver.resources.coverages.models.EOObject>`.

        :param queryset: the ``QuerySet`` to filter
        :param containment: either "overlaps" or "contains"
        :returns: a ``QuerySet`` with additional filters applied
        """

        if not len(self):
            return queryset

        filters = self.get_filters(containment)
        return queryset.filter(**filters)

    def matches(self, eo_object, containment="overlaps"):
        """ Check if the given :class:`EOObject
        <eoxserver.resources.coverages.models.EOObject>` matches the given
        subsets.

        :param eo_object: the ``EOObject`` to match
        :param containment: either "overlaps" or "contains"
        :returns: a boolean value indicating if the object is contained in the
                  given subsets
        """
        if not len(self):
            return True

        bbox = [None, None, None, None]
        srid = self.srid
        if srid is None:
            srid = 4326
        max_extent = crss.crs_bounds(srid)
        tolerance = crss.crs_tolerance(srid)

        # check if time intervals are configured as "open" or "closed"
        config = get_eoxserver_config()
        reader = SubsetConfigReader(config)
        # note that the operator is inverted from filter() above as the
        # filters use an inclusive search whereas here it's exclusive
        if reader.time_interval_interpretation == "closed":
            gt_op = operator.gt
            lt_op = operator.lt
        else:
            gt_op = operator.ge
            lt_op = operator.le

        footprint = eo_object.footprint
        begin_time = eo_object.begin_time
        end_time = eo_object.end_time

        for subset in self:
            if isinstance(subset, Slice):
                is_slice = True
                value = subset.value
            elif isinstance(subset, Trim):
                is_slice = False
                low = subset.low
                high = subset.high
                # we need the value in case low == high
                value = low

            if subset.is_temporal:
                if is_slice or (low == high and containment == "overlaps"):
                    if begin_time > value or end_time < value:
                        return False
                elif low == high:
                    if begin_time < value or end_time > value:
                        return False
                else:
                    # check if the temporal bounds must be strictly contained
                    if containment == "contains":
                        if high is not None:
                            if gt_op(end_time, high):
                                return False
                        if low is not None:
                            if lt_op(begin_time, low):
                                return False
                    # or just overlapping
                    else:
                        if high is not None:
                            if gt_op(begin_time, high):
                                return False
                        if low is not None:
                            if lt_op(end_time, low):
                                return False

            else:
                if is_slice:
                    if subset.is_x:
                        line = LineString(
                            (value, max_extent[1]),
                            (value, max_extent[3])
                        )
                    else:
                        line = LineString(
                            (max_extent[0], value),
                            (max_extent[2], value)
                        )
                    line.srid = srid
                    if srid != 4326:
                        line.transform(4326)

                    if not line.intersects(footprint):
                        return False

                else:
                    if subset.is_x:
                        bbox[0] = subset.low
                        bbox[2] = subset.high
                    else:
                        bbox[1] = subset.low
                        bbox[3] = subset.high

        if bbox != [None, None, None, None]:
            bbox = map(
                lambda v: v[0] if v[0] is not None else v[1],
                zip(bbox, max_extent)
            )

            bbox[0] -= tolerance
            bbox[1] -= tolerance
            bbox[2] += tolerance
            bbox[3] += tolerance

            logger.debug(
                "Applying BBox %s with containment '%s'." % (bbox, containment)
            )

            poly = Polygon.from_bbox(bbox)
            poly.srid = srid

            if srid != 4326:
                poly.transform(4326)
            if containment == "overlaps":
                if not footprint.intersects(poly):
                    return False
            elif containment == "contains":
                if not footprint.within(poly):
                    return False
        return True

    def _check_subset(self, subset):
        if not isinstance(subset, Subset):
            raise ValueError("Supplied argument is not a subset.")

        if not isinstance(subset, self.allowed_types):
            raise InvalidSubsettingException(
                "Supplied subset is not allowed."
            )

        if self.has_x and subset.is_x:
            raise InvalidSubsettingException(
                "Multiple subsets for X-axis given."
            )

        if self.has_y and subset.is_y:
            raise InvalidSubsettingException(
                "Multiple subsets for Y-axis given."
            )

        if self.has_t and subset.is_temporal:
            raise InvalidSubsettingException(
                "Multiple subsets for time-axis given."
            )

    @property
    def xy_bbox(self):
        """ Returns the minimum bounding box for all X and Y subsets.

        :returns: a list of four elements [minx, miny, maxx, maxy], which might
                  be ``None``
        """
        bbox = [None, None, None, None]
        for subset in self:
            if subset.is_x:
                if isinstance(subset, Trim):
                    bbox[0] = subset.low
                    bbox[2] = subset.high
                else:
                    bbox[0] = bbox[2] = subset.value
            elif subset.is_y:
                if isinstance(subset, Trim):
                    bbox[1] = subset.low
                    bbox[3] = subset.high
                else:
                    bbox[1] = bbox[3] = subset.value

        return bbox

    def bounding_polygon(self, coverage):
        """ Returns a minimum bounding :class:`django.contrib.gis.geos.Polygon`
        for the given :class:`Coverage
        <eoxserver.render.coverages.objects.Coverage>`

        :param coverage: the coverage to calculate the bounding polygon for
        :returns: the calculated ``Polygon``
        """

        srid = SpatialReference(coverage.grid.coordinate_reference_system).srid
        extent = coverage.extent
        size_x, size_y = coverage.size
        footprint = coverage.footprint

        subset_srid = self.srid

        if subset_srid is None:
            bbox = list(extent)
        else:
            bbox = list(footprint.extent)

        for subset in self:
            if not isinstance(subset, Trim) or subset.is_temporal:
                continue

            if subset_srid is None:
                # transform coordinates from imageCRS to coverages CRS
                if subset.is_x:
                    if subset.low is not None:
                        l = max(float(subset.low) / float(size_x), 0.0)
                        bbox[0] = extent[0] + l * (extent[2] - extent[0])

                    if subset.high is not None:
                        l = max(float(subset.high) / float(size_x), 0.0)
                        bbox[2] = extent[0] + l * (extent[2] - extent[0])

                elif subset.is_y:
                    if subset.low is not None:
                        l = max(float(subset.low) / float(size_y), 0.0)
                        bbox[1] = extent[3] - l * (extent[3] - extent[1])

                    if subset.high is not None:
                        l = max(float(subset.high) / float(size_y), 0.0)
                        bbox[3] = extent[3] - l * (extent[3] - extent[1])

            else:
                if subset.is_x:
                    if subset.low is not None:
                        bbox[0] = max(subset.low, bbox[0])

                    if subset.high is not None:
                        bbox[2] = min(subset.high, bbox[2])

                if subset.is_y:
                    if subset.low is not None:
                        bbox[1] = max(subset.low, bbox[1])

                    if subset.high is not None:
                        bbox[3] = min(subset.high, bbox[3])

        if subset_srid is None:
            poly = Polygon.from_bbox(bbox)
            poly.srid = srid

        else:
            poly = Polygon.from_bbox(bbox)
            poly.srid = subset_srid

        return poly


class Subset(object):
    """ Base class for all subsets.
    """

    def __init__(self, axis):
        axis = axis.lower()
        if axis not in all_axes:
            raise InvalidAxisLabelException(axis)
        self.axis = axis

    @property
    def is_temporal(self):
        return self.axis in temporal_axes

    @property
    def is_x(self):
        return self.axis in x_axes

    @property
    def is_y(self):
        return self.axis in y_axes


class Slice(Subset):
    """ Slice subsets reduce the dimension of the subsetted object by one and
    slice the given ``axis`` at the specified ``value``.

    :param axis: the axis name
    :param value: the slice point
    """
    def __init__(self, axis, value):
        super(Slice, self).__init__(axis)
        self.value = value

    def __repr__(self):
        return "Slice: %s[%s]" % (self.axis, self.value)


class Trim(Subset):
    """ Trim subsets reduce the domain of the specified ``axis``

    :param axis: the axis name
    :param low: the lower end of the ``Trim``; if omitted, the ``Trim`` has no
                lower bound
    :param high: the upper end of the ``Trim``; if omitted, the ``Trim`` has no
                 upper bound
    """
    def __init__(self, axis, low=None, high=None):
        super(Trim, self).__init__(axis)

        if low is not None and high is not None and low > high:
            raise InvalidSubsettingException(
                "Invalid bounds: lower bound greater than upper bound."
            )

        self.low = low
        self.high = high

    def __repr__(self):
        return "Trim: %s[%s:%s]" % (
            self.axis, self.low, self.high
        )


temporal_axes = ("t", "time", "phenomenontime")
x_axes = ("x", "lon", "long")
y_axes = ("y", "lat")
z_axes = ("z", "height")
all_axes = temporal_axes + x_axes + y_axes + z_axes


def is_temporal(axis):
    """ Returns whether or not an axis is a temporal one.
    """
    return (axis.lower() in temporal_axes)


class SubsetConfigReader(config.Reader):
    section = "services.owscommon"
    time_interval_interpretation = config.Option(
        default="closed", type=enum(("closed", "open"), False)
    )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.db import models
from django.core.exceptions import ValidationError
from django.utils.encoding import python_2_unicode_compatible

from eoxserver.backends.storages import get_handler_class_by_name


optional = dict(null=True, blank=True)
mandatory = dict(null=False, blank=False)


# ==============================================================================
# Models
# ==============================================================================


@python_2_unicode_compatible
class StorageAuth(models.Model):
    """ Model to symbolize authorization for storages.
    """
    url = models.CharField(max_length=1024, **mandatory)
    storage_auth_type = models.CharField(max_length=32, **mandatory)
    name = models.CharField(max_length=1024, null=True, blank=True, unique=True)
    auth_parameters = models.TextField()

    def __str__(self):
        return "%s: %s" % (self.storage_auth_type, self.url)

    def clean(self):
        validate_storage_auth(self)


@python_2_unicode_compatible
class Storage(models.Model):
    """ Model to symbolize storages that provide file or other types of access
        to data items.
    """
    url = models.CharField(max_length=1024, **mandatory)
    storage_type = models.CharField(max_length=32, **mandatory)
    name = models.CharField(max_length=1024, null=True, blank=True, unique=True)
    storage_auth = models.ForeignKey(StorageAuth, on_delete=models.CASCADE, **optional)

    parent = models.ForeignKey("self", on_delete=models.CASCADE, **optional)

    def __str__(self):
        return "%s: %s" % (self.storage_type, self.url)

    def clean(self):
        validate_storage(self)


@python_2_unicode_compatible
class DataItem(models.Model):
    """ Abstract model for locateable data items contributing to a dataset.
    """

    storage = models.ForeignKey(Storage, on_delete=models.CASCADE, **optional)
    location = models.CharField(max_length=1024, **mandatory)
    format = models.CharField(max_length=64, **optional)

    class Meta:
        abstract = True

    def __str__(self):
        if self.format:
            return "%s (%s)" % (self.location, self.format)
        return self.location


# ==============================================================================
# Validators
# ==============================================================================


def validate_storage(storage):
    parent = storage.parent

    handler = get_handler_class_by_name(storage.storage_type)
    if not handler:
        raise ValidationError(
            'Storage type %r is not supported.' % storage.storage_type
        )

    if parent:
        parent_handler = get_handler_class_by_name(parent.storage_type)
        if not handler.allows_parent_storage:
            raise ValidationError(
                'Storage type %r does not allow parent storages'
                % storage.storage_type
            )
        elif not parent_handler.allows_child_storages:
            raise ValidationError(
                'Parent storage type %r does not allow child storages'
                % parent.storage_type
            )

    while parent:
        if parent == storage:
            raise ValidationError('Circular reference detected')
        parent = parent.parent


def validate_storage_auth(storage_auth):
    pass



# -*- coding: utf-8 -*-
# Generated by Django 2.2.9 on 2020-08-18 13:57
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0007_typemodels'),
    ]

    operations = [
        migrations.AddField(
            model_name='ProductMetadata',
            name='across_track_incidence_angle',
            field=models.FloatField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name='ProductMetadata',
            name='along_track_incidence_angle',
            field=models.FloatField(blank=True, db_index=True, null=True),
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core.decoders import config, typelist, Choice


class CapabilitiesConfigReader(config.Reader):
    section = "services.ows"

    update_sequence     = config.Option(default="0")
    name                = config.Option(default="None")
    title               = config.Option(default="None")
    abstract            = config.Option(default="None")
    keywords            = config.Option(type=typelist(str, ","), default=[])
    fees                = config.Option(default="None")
    access_constraints  = config.Option(default="None")
    provider_name       = config.Option(default="None")
    provider_site       = config.Option(default="None")
    individual_name     = config.Option(default="None")
    position_name       = config.Option(default="None")
    phone_voice         = config.Option(default="None")
    phone_facsimile     = config.Option(default="None")
    delivery_point      = config.Option(default="None")
    city                = config.Option(default="None")
    administrative_area = config.Option(default="None")
    postal_code         = config.Option(default="None")
    country             = config.Option(default="None")
    electronic_mail_address = config.Option(default="None")
    onlineresource      = config.Option(default="None")
    hours_of_service    = config.Option(default="None")
    contact_instructions = config.Option(default="None")
    role                = config.Option(default="None")

    http_service_url = Choice(
        config.Option("http_service_url", section="services.owscommon", required=True),
        config.Option("http_service_url", section="services.ows", required=True),
    )


class WCSEOConfigReader(config.Reader):
    section = "services.ows.wcs20"
    paging_count_default = config.Option(type=int, default=None)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013-2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

class OWS10Exception(Exception):
    """ Base OWS 1.0 exception of the WPS 1.0.0 exceptions """
    http_status_code = 400
    def __init__(self, code, locator, message):
        self.code = code
        self.locator = locator
        Exception.__init__(self, message)

#-------------------------------------------------------------------------------
# All possible WPS 1.0.0 exceptions. For list of OWS exception used by WPS
# see OGC 05-007r7 Table 38 and Table 62

class NoApplicableCode(OWS10Exception):
    http_status_code = 500
    def __init__(self, message, locator=None):
        OWS10Exception.__init__(self, "NoApplicableCode", locator, message)

class MissingParameterValue(OWS10Exception):
    def __init__(self, message, locator):
        OWS10Exception.__init__(self, "MissingParameterValue", locator, message)

class InvalidParameterValue(OWS10Exception):
    def __init__(self, message, locator):
        OWS10Exception.__init__(self, "InvalidParameterValue", locator, message)

class NotEnoughStorage(OWS10Exception):
    http_status_code = 507
    def __init__(self, message):
        OWS10Exception.__init__(self, "NotEnoughStorage", None, message)

class ServerBusy(OWS10Exception):
    http_status_code = 503
    def __init__(self, message):
        OWS10Exception.__init__(self, "ServerBusy", None, message)

class FileSizeExceeded(OWS10Exception):
    def __init__(self, message, locator):
        OWS10Exception.__init__(self, "FileSizeExceeded", locator, message)

class StorageNotSupported(OWS10Exception):
    def __init__(self, message):
        OWS10Exception.__init__(self, "StorageNotSupported", None, message)

class VersionNegotiationFailed(OWS10Exception):
    def __init__(self, message, locator):
        OWS10Exception.__init__(self, "VersionNegotiationFailed", locator, message)

#-------------------------------------------------------------------------------
# Derived specific exceptions.
#
# Note that WPS 1.0.0 allows use of "vendor specific exception code" as locator
# for the default "NoApplicableCode" exceptions.

class NoSuchProcessError(InvalidParameterValue):
    def __init__(self, identifier):
        msg = "No such process: %s" % identifier
        InvalidParameterValue.__init__(self, msg, "process identifier")

class InvalidOutputError(InvalidParameterValue):
    def __init__(self, output_id):
        message = "Invalid output '%s'!"%(output_id)
        InvalidParameterValue.__init__(self, message, output_id)

class InvalidOutputValueError(NoApplicableCode):
    def __init__(self, output_id, message=""):
        message = "Invalid output value of '%s'! %s"%(output_id, message)
        NoApplicableCode.__init__(self, message, output_id)

class InvalidOutputDefError(InvalidParameterValue):
    def __init__(self, output_id, message=""):
        message = "Invalid output definition of '%s'! %s"%(output_id, message)
        InvalidParameterValue.__init__(self, message, output_id)

class InvalidInputError(InvalidParameterValue):
    def __init__(self, input_id):
        message = "Invalid input '%s'!"%(input_id)
        InvalidParameterValue.__init__(self, message, input_id)

class InvalidInputValueError(InvalidParameterValue):
    def __init__(self, input_id, message=""):
        message = "Invalid input value of '%s'! %s"%(input_id, message)
        InvalidParameterValue.__init__(self, message, input_id)

class InvalidInputReferenceError(InvalidParameterValue):
    def __init__(self, input_id, message=""):
        message = "Invalid input '%s' reference! %s"%(input_id, message)
        InvalidParameterValue.__init__(self, message, input_id)

class MissingRequiredInputError(InvalidParameterValue):
    def __init__(self, input_id):
        message = "Missing required input '%s'!"%(input_id)
        InvalidParameterValue.__init__(self, message, input_id)

class ExecuteError(NoApplicableCode):
    def __init__(self, message="", locator="process.execute()"):
        NoApplicableCode.__init__(self, message, locator)



#-------------------------------------------------------------------------------
#
# WPS 1.0 ProcessDescriptsions XML encoders
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.services.ows.wps.v10.util import (
    OWS, WPS, NIL, ns_wps, ns_xlink, ns_xml
)
from .parameters import encode_input_descr, encode_output_descr
from .base import WPS10BaseXMLEncoder
from eoxserver.services.ows.wps.parameters import fix_parameter


class WPS10ProcessDescriptionsXMLEncoder(WPS10BaseXMLEncoder):
    """ WPS 1.0 ProcessDescriptions XML response encoder. """

    @staticmethod
    def encode_process_descriptions(processes):
        """ Encode the ProcessDescriptions XML document. """
        _proc = [encode_process_full(p) for p in processes]
        _attr = {
            "service": "WPS",
            "version": "1.0.0",
            ns_xml("lang"): "en-US",
        }
        return WPS("ProcessDescriptions", *_proc, **_attr)


def encode_process_brief(process):
    """ Encode a brief process description (Process element) of the
    Capabilities XML document.
    """
    return _encode_process_brief(process, WPS("Process"))


def encode_process_full(process):
    """ Encode a full process description (ProcessDescription element) of the
    ProcessDescriptions XML document.
    """
    if getattr(process, 'asynchronous', False):
        supports_store = True
        supports_update = True
    else:
        supports_store = False
        supports_update = False

    # TODO: remove backward compatibility support for inputs/outputs dicts
    if isinstance(process.inputs, dict):
        process.inputs = process.inputs.items()
    if isinstance(process.outputs, dict):
        process.outputs = process.outputs.items()

    inputs = [
        item for item in (
            encode_input_descr(fix_parameter(n, p)) for n, p in process.inputs
        ) if item is not None
    ]
    outputs = [
        encode_output_descr(fix_parameter(n, p)) for n, p in process.outputs
    ]

    elem = _encode_process_brief(process, NIL("ProcessDescription"))
    if supports_store:
        elem.attrib["storeSupported"] = "true"
    if supports_update:
        elem.attrib["statusSupported"] = "true"
    elem.append(NIL("DataInputs", *inputs))
    elem.append(NIL("ProcessOutputs", *outputs))

    return elem


def _encode_process_brief(process, elem):
    """ Insert a brief process description into an XML element passed as the
    second argument.
    The brief process description is shared by both the Capabilities and
    ProcessDescriptions XML encoders.
    """
    identifier = getattr(process, 'identifier', type(process).__name__)
    title = getattr(process, 'title', identifier)
    abstract = getattr(process, 'description', process.__doc__)
    version = getattr(process, "version", "1.0.0")
    metadata = getattr(process, "metadata", {})
    profiles = getattr(process, "profiles", [])
    wsdl = getattr(process, "wsdl", None)

    elem.append(OWS("Identifier", identifier))
    elem.append(OWS("Title", title))
    elem.attrib[ns_wps("processVersion")] = version
    if abstract:
        elem.append(OWS("Abstract", abstract))
    elem.extend(_encode_metadata(k, metadata[k]) for k in metadata)
    elem.extend(WPS("Profile", profile) for profile in profiles)
    if wsdl:
        elem.append(WPS("WSDL", **{ns_xlink("href"): wsdl}))

    return elem


def _encode_metadata(title, href):
    """ Encode one Metadata element. """
    return OWS("Metadata", **{ns_xlink("title"): title, ns_xlink("href"): href})



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-11-10 12:22
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import eoxserver.resources.coverages.models
import re


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='browsetype',
            name='alpha_nodata_value',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='alpha_range_max',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='alpha_range_min',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='blue_nodata_value',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='blue_range_max',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='blue_range_min',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='green_nodata_value',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='green_range_max',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='green_range_min',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='red_or_grey_nodata_value',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='red_or_grey_range_max',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='red_or_grey_range_min',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='browsetype',
            name='alpha_expression',
            field=models.CharField(blank=True, max_length=512, null=True, validators=[eoxserver.resources.coverages.models.band_expression_validator]),
        ),
        migrations.AlterField(
            model_name='browsetype',
            name='blue_expression',
            field=models.CharField(blank=True, max_length=512, null=True, validators=[eoxserver.resources.coverages.models.band_expression_validator]),
        ),
        migrations.AlterField(
            model_name='browsetype',
            name='green_expression',
            field=models.CharField(blank=True, max_length=512, null=True, validators=[eoxserver.resources.coverages.models.band_expression_validator]),
        ),
        migrations.AlterField(
            model_name='browsetype',
            name='name',
            field=models.CharField(blank=True, max_length=256, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')]),
        ),
        migrations.AlterField(
            model_name='browsetype',
            name='red_or_grey_expression',
            field=models.CharField(blank=True, max_length=512, null=True, validators=[eoxserver.resources.coverages.models.band_expression_validator]),
        ),
    ]



#-------------------------------------------------------------------------------
#
# WPS 1.0 Capabilities XML encoders
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
#pylint: disable=bad-continuation

from eoxserver.core.config import get_eoxserver_config
from eoxserver.services.ows.dispatch import filter_handlers
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.ows.wps.v10.util import (
    OWS, WPS, ns_xlink, ns_xml,
)
from .process_description import encode_process_brief
from .base import WPS10BaseXMLEncoder


class WPS10CapabilitiesXMLEncoder(WPS10BaseXMLEncoder):
    """ WPS 1.0 Capabilities XML response encoder. """
    @staticmethod
    def encode_capabilities(processes):
        """ Encode Capabilities XML document. """
        conf = CapabilitiesConfigReader(get_eoxserver_config())

        # Avoid duplicate process offerings ...
        process_set = set()
        process_offerings = []
        for process in processes:
            process_identifier = (
                getattr(process, 'identifier', None) or type(process).__name__
            )
            if process_identifier not in process_set:
                process_offerings.append(encode_process_brief(process))
                process_set.add(process_identifier)

        return WPS("Capabilities",
            OWS("ServiceIdentification",
                OWS("Title", conf.title),
                OWS("Abstract", conf.abstract),
                OWS("Keywords", *(OWS("Keyword", kw) for kw in conf.keywords)),
                OWS("ServiceType", "WPS"),
                OWS("ServiceTypeVersion", "1.0.0"),
                OWS("Fees", conf.fees),
                OWS("AccessConstraints", conf.access_constraints),
            ),
            OWS("ServiceProvider",
                OWS("ProviderName", conf.provider_name),
                OWS("ProviderSite", **{ns_xlink("href"): conf.provider_site}),
                OWS("ServiceContact",
                    OWS("IndividualName", conf.individual_name),
                    OWS("PositionName", conf.position_name),
                    OWS("ContactInfo",
                        OWS("Phone",
                            OWS("Voice", conf.phone_voice),
                            OWS("Facsimile", conf.phone_facsimile)
                        ),
                        OWS("Address",
                            OWS("DeliveryPoint", conf.delivery_point),
                            OWS("City", conf.city),
                            OWS("AdministrativeArea", conf.administrative_area),
                            OWS("PostalCode", conf.postal_code),
                            OWS("Country", conf.country),
                            OWS(
                                "ElectronicMailAddress",
                                conf.electronic_mail_address
                            )
                        )
                    )
                )
            ),
            _encode_operations_metadata(conf),
            WPS("ProcessOfferings", *process_offerings),
            WPS("Languages",
                WPS("Default",
                    OWS("Language", "en-US")
                ),
                WPS("Supported",
                    OWS("Language", "en-US")
                )
            ),
            # TODO: WPS("WSDL")
            **{
                "service": "WPS",
                "version": "1.0.0",
                ns_xml("lang"): "en-US",
                "updateSequence": conf.update_sequence,
            }
        )


def _encode_operations_metadata(conf):
    """ Encode OperationsMetadata XML element. """
    versions = ("1.0.0",)
    get_handlers = filter_handlers(
        service="WPS", versions=versions, method="GET"
    )
    post_handlers = filter_handlers(
        service="WPS", versions=versions, method="POST"
    )
    all_handlers = sorted(
        set(get_handlers + post_handlers), key=lambda h: h.request
    )
    url = conf.http_service_url
    return OWS("OperationsMetadata", *[
        OWS("Operation",
            OWS("DCP",
                OWS("HTTP",
                    OWS("Get", **{ns_xlink("href"): url}),
                    OWS("Post", **{ns_xlink("href"): url}),
                )
            ), name=handler.request
        )
        for handler in all_handlers
    ])
